#!/usr/bin/env python
from __future__ import print_function
from diapysef.pasefdata import PasefData
from diapysef.pasefdata import PasefMQData
import sys

"""
Annotate the output of a Maxquant analysis with ion mobility in 1/K0.
Usage: annotate_mq_ionmobility.py mqout_dir pasef_analysis_dir output_prefix
"""

if len(sys.argv) < 4:
    print("Usage: annotate_mq_ionmobility.py mqout_dir pasef_analysis_dir output_prefix")
    sys.exit()

mqout = sys.argv[1]
pasefdata = sys.argv[2]
outfile = sys.argv[3]

pas = PasefData(pasefdata)
mq = PasefMQData(mqout)

mq.get_all_peptides(pas)
mq.get_evidence(pas)

mq.all_peptides.to_csv(outfile + "_all_peptides.csv")
mq.evidence.to_csv(outfile + "_evidence.csv")



#!/usr/bin/env python
from __future__ import print_function
import pyopenms
import numpy as np


class MergeConsumer():
    """
        Merging consumer that merges MS2 spectra with the same precursor. The
        number of consecutive spectra to be merged is a user parameter.

        This class merges m/z and intensity coordinates as well as the first
        FloatDataArray.

        Does not implement chromatogram consuming.
    """

    def __init__(self, consumer, merge_nr):
        self._internal_consumer = consumer
        self._spectrum_storage = {}
        self._merge_nr = merge_nr

    def __del__(self):
        """
        Cleanup: write all stored spectra to disk
        """
        for k, v in self._spectrum_storage.items():
            if v:
                merge_spec = self._mergeSpectra(v)
                self._internal_consumer.consumeSpectrum(merge_spec)

    def consumeSpectrum(self, s):
        """
            consume individual spectra:
                - write MS1 spectra directly to disk
                - collect MS2 spectra and write afer merging n spectra together
        """
        if s.getMSLevel() == 1:
            self._internal_consumer.consumeSpectrum(s)
        else:

            mz = int(s.getPrecursors()[0].getMZ()*10)
            tmp = self._spectrum_storage.get(mz, [])
            tmp.append(s)

            if len(tmp) >= self._merge_nr:
                merge_spec = self._mergeSpectra(tmp)
                self._internal_consumer.consumeSpectrum(merge_spec)
                tmp = []

            self._spectrum_storage[mz] = tmp

    def _mergeSpectra(self, tmp):
        """
        Perform spectral merging
         - we pick one spectrum (the first one) as our reference and add
           all other data from the other spectra to it
         - we check that merged spectra have equal precursor m/z and same float array
        """

        merge_spec = pyopenms.MSSpectrum(tmp[0])
        fda = tmp[0].getFloatDataArrays()[0]
        fda.clear()
        allmz = []
        allint = []
        for q in tmp:
            m, i = q.get_peaks()
            allmz.append(m)
            allint.append(i)

            # Sanity checks, precursors of merged spectra and float arrays need to match
            assert q.getPrecursors()[0].getMZ() - merge_spec.getPrecursors()[0].getMZ() < 1e-5
            assert len(q.getFloatDataArrays()) == len(merge_spec.getFloatDataArrays())
            assert q.getFloatDataArrays()[0].getName() == merge_spec.getFloatDataArrays()[0].getName()

            # TODO this is not very efficient, fix in pyOpenMS!
            for d in q.getFloatDataArrays()[0]:
                fda.push_back(d)

        mz = np.concatenate(allmz)
        intens = np.concatenate(allint)

        # create merge spec
        merge_spec.set_peaks( (mz, intens) )
        merge_spec.setFloatDataArrays([fda])
        merge_spec.sortByPosition()
        return merge_spec




#!/usr/bin/env python
from __future__ import print_function
from diapysef.plotting import plot_window_layout
import sys
import pandas as pd
"""
Plots the window layout of a tims experiment with precursor density of a pasef experiment.
"""
if len(sys.argv) < 3:
    print("Usage: plot_dia_windows.py window_file precursor_map")
    sys.exit()

windows = pd.read_csv(sys.argv[1])
precursors = pd.read_csv(sys.argv[2])

plot_window_layout(windows = windows, precursor_map = precursors)



#!/usr/bin/env python
from __future__ import print_function
import pickle
import os.path
import pandas as pd
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.patches as patches
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import matplotlib
from tqdm import tqdm
# Logging
import logging
logging.getLogger('matplotlib').setLevel(logging.WARNING)

# Plotting
matplotlib.use('Agg')

# Data


def plot_window_layout(windows, precursor_map=None, display_sc=False):
    """Plots the windows with an optional background of ms1 features"""
    # if type(precursor_map) == 'PasefMQData':
    #     if(not hasattr(precursor_map, all_peptides)):
    #         precursor_map.get_all_peptides()
    #         precursor_map.annotate_ion_mobility()
    #     mq = precursor_map.all_peptides
    # elif type(precursor_map) == 'DataFrame':
    w = windows

    if precursor_map is None:
        mq = pd.read_pickle(os.path.join(os.path.dirname(
            __file__), 'data/evidence_example.pickle'))
        # ax = pickle.load(open(os.path.join(os.path.dirname(__file__), 'data/all_peptides_density.pickle'), 'rb'))
    else:
        mq = precursor_map
    if not display_sc:
        mq = mq[mq.Charge > 1]

    f, ax = plt.subplots(figsize=(8, 6))
    ax.hist2d(mq['m/z'], mq['IonMobilityIndexK0'],
              bins=[1000, 1000], norm=LogNorm())

    for p in [
            patches.Rectangle(
                (w['IsolationMz'][i]-w['IsolationWidth'][i]/2, w['IMend'][i]),
                w['IsolationWidth'][i],
                w['IMstart'][i] - w['IMend'][i],
                alpha=0.4) for i in range(len(w))
    ]:
        ax.add_patch(p)

    ax.set(xlim=(min(mq['m/z'].min(), (w['IsolationMz'] - w['IsolationWidth']/2).min()),
                 max(mq['m/z'].max(), (w['IsolationMz'] + w['IsolationWidth']/2).max())),
           ylim=(min(mq['IonMobilityIndexK0'].min(), w['IMend'].min()),
                 max(mq['IonMobilityIndexK0'].max(), w['IMstart'].max())))
    plt.xlabel('m/z')
    plt.ylabel('Ion Mobility (1/K0)')

    plt.show()


def get_2d_heatmap_data(data_long, x_col='rt', y_col='im', z_col='int'):
    """
    Get X, Y and Z data for plotting a heatmap

    Params:
        data_long: (data.frame) data.frame containing data for plotting, must have at least 3 columns to extract data for X, Y and Z
        x_col: (str) the column that should be used for the X-axis
        y_col: (str) the column that should by used for the Y-axis
        z_col: (str) the column that should be used for the Z-axis

    Returns:
        Returns X, Y and Z values of equal dimensions for plotting to grid on a heatmap
    """

    x = data_long[x_col].to_numpy()
    y = data_long[y_col].to_numpy()
    z = data_long[z_col].to_numpy()

    # print(f"X: {x.shape} | Y: {y.shape} | Z: {z.shape}")

    # Pivot table to grid
    pdata = pd.DataFrame(data={'x': x, 'y': y, 'z': z})
    pdata = pdata.pivot_table(index='y', columns='x', values='z')

    X = pdata.columns.to_numpy()
    Y = pdata.index.to_numpy()
    Z = pdata.to_numpy()

    return X, Y, Z


def plot_2d_rt_im_heatmap(data, current_peptide, plot_contours=False, fig=plt.figure(1)):
    """
    Plot a Heatmap of RT and IM

    Params:
        data: (data.frame) data.frame containing ms_level information, retention time, ion mobility and intensity
        current_peptide: (str) string of current peptide being plotting
        plot_contours: (bool) Should contour lines be plotted?
        fig: (matplotlib figure object) figure object from matplotlib pyplot

    Returns:
        None
    """
    plt.suptitle(f"Peptide(z): {current_peptide}")
    # Plot MS1 Data if available
    data_sub = data.loc[(data.ms_level == 1)]
    if data_sub.shape[0] != 0:
        ax1 = plt.subplot(1, 2, 1)
        # Get axis data
        X, Y, Z = get_2d_heatmap_data(data_sub)
        # Plot heatmap
        c = plt.pcolormesh(X, Y, Z, cmap='afmhot_r')
        if plot_contours:
            ax1.contour(X, Y, Z, colors='red', alpha=0.55)
        ax1.set_title('MS-Level = 1')
        fig.colorbar(c, ax=ax1)
    # Plot MS2 Data if available
    data_sub = data.loc[(data.ms_level == 2)]
    if data_sub.shape[0] != 0:
        ax2 = plt.subplot(1, 2, 2)
        # Get axis data
        X, Y, Z = get_2d_heatmap_data(data_sub)
        # Plot heatmap
        c1 = plt.pcolormesh(X, Y, Z, cmap='afmhot_r')
        if plot_contours:
            ax2.contour(X, Y, Z, colors='red', alpha=0.55)
        ax2.set_title('MS-Level = 2')
        fig.colorbar(c1, ax=ax2)
    plt.tight_layout(rect=[0, 0.03, 1, 0.95])
    fig.add_subplot(111, frameon=False)
    plt.tick_params(labelcolor='none', which='both', top=False,
                    bottom=False, left=False, right=False)
    plt.xlabel("Retention Time [sec]")
    plt.ylabel("Ion Mobility [1/K0]")


def save_report_2d_rt_im_heatmap(infile, outpdf="diapasef_rt_im_heatmap.pdf", plot_contours=False):
    """
    Save a report of 2D RT and IM heatmap plots to pdf

    params:
        infile: (str) tsv file containing data from targeted extracted export experiment. Should contain information for peptide, charge, ms_level, retention time, ion mobility and intensity
        outpdf: (str) the pdf file name to save the plots to
        plot_contours: (bool) Should contour lines be plotted?

    Returns:
        None
    """
    data = pd.read_csv(infile, sep="\t")
    # Add a group_id to group peptide and charge
    data['group_id'] = data['peptide'] + '_' + data['charge'].astype(str)

    unique_peptide_groups = np.unique(data[['group_id']])
    with PdfPages(outpdf) as pdf:
        pbar = tqdm(range(len(unique_peptide_groups)))
        pbar_desc = "INFO: Plotting"
        fig = plt.figure(1, figsize=(12.75, 8.25))
        for peptide_group in pbar:
            current_peptide = unique_peptide_groups[peptide_group]

            data_peptide_sub = data.loc[(
                data.group_id.isin([current_peptide]))]

            pbar_desc = f"INFO: Plotting..{current_peptide}"
            pbar.set_description(pbar_desc)
            plot_2d_rt_im_heatmap(
                data_peptide_sub, current_peptide, plot_contours, fig)

            # Save to PDF
            pdf.savefig()  # save on the fly
            plt.clf()  # clear figure once saved
        plt.close()



#!/usr/bin/env python
from __future__ import print_function

"""Conversion program to convert a Bruker TIMS file to a single mzML

"""
import argparse
import sys
import sqlite3
import time
import pyopenms
import numpy as np
#import matplotlib.pyplot as plt
from ctypes import cdll
from tqdm import tqdm

from .timsdata import TimsData
from .merge_consumer import MergeConsumer
from .splitting_consumer import SplittingConsumer
from .util import setCompressionOptions


try:
    if __name__ == "__main__":
        if sys.platform[:5] == "win32" or sys.platform[:5] == "win64":
            libname = "timsdata.dll"
            dll = cdll.LoadLibrary(libname)
        elif sys.platform[:5] == "linux":
            libname = "libtimsdata.so"
            dll = cdll.LoadLibrary(libname)
        else:
            raise Exception("Unsupported platform.")
except OSError as e:
    print("This functionality can only be carried out if the bruker sdk is present. Please install it first. The sdk can be installed by installing proteowizard(version >=3, http://proteowizard.sourceforge.net), or by placing the a library file in your path (For windows this will be timsdata.dll and for Linux libtimsdata.so).\n")
    sys.exit()

def store_frame(frame_id, td, conn, exp, verbose=False, compressFrame=True, keep_frames=False):
    """
    Store a single frame as an individual mzML file

    Note that there are two ways to store the data:
      (i) Multiple spectra per frame (for visualization), compressFrame is False. This is the
      easiest way to visualize and process the data but involves a few hacks,
      namely storing the IM axis as the RT of each spectrum.

      (ii) One spectrum per frame, compressFrame is True. This puts all peaks
      into a single spectrum (while storing the IM data in an extra array).
      This is more efficient for storage and allows analysis that is ignorant
      of the IM dimension.

      Note that msms = 2 means that we have an MS2 scan whereas msms = 8 stands
      for pasef scan. (New tdf 5.1 has msms = 9 for pasef scan)
    """
    # Get a projected mass spectrum:
    q = conn.execute("SELECT NumScans, Time, Polarity, MsMsType FROM Frames WHERE Id={0}".format(frame_id))
    tmp = q.fetchone()
    num_scans = tmp[0]
    time = tmp[1]
    pol = tmp[2]
    msms = int(tmp[3])

    center = -1
    width = -1
    next_scan_switch = -1
    mslevel = 1
    scan_data = []
    scan_data_it = 0
    scanBoundariesk0 = (0.0, 0.0)

    in_scan = False
    scandata = None

    # Check whether we have a MS2 or a PASEF scan
    if msms == 2:
        q = conn.execute("SELECT TriggerMass, IsolationWidth, PrecursorCharge, CollisionEnergy FROM FrameMsMsInfo WHERE Frame={0}".format(frame_id))
        tmp = q.fetchone()
        center = float(tmp[0])
        width = float(tmp[1])
        mslevel = 2
    # new tdf 5.1 has pasef scan msms = 9 
    elif msms == 9:
        q = conn.execute("SELECT IsolationMz, IsolationWidth, ScanNumBegin, ScanNumEnd, CollisionEnergy, Frame FROM DiaFrameMsMsWindows INNER JOIN DiaFrameMsMsInfo ON DiaFrameMsMsWindows.WindowGroup = DiaFrameMsMsInfo.WindowGroup WHERE Frame={0} ORDER BY  ScanNumBegin DESC".format(frame_id))
        scandata = q.fetchall()
        tmp = scandata[scan_data_it]
        center = float(tmp[0])
        width = float(tmp[1])
        scan_start = int(tmp[2])
        scan_end = int(tmp[3])
        next_scan_switch = scan_end
        scanBoundariesk0 = td.scanNumToOneOverK0(frame_id, np.array([scan_end, scan_start])) #spectrum 1/k0 boundaries for the first swath in the frame
        # Check if we already are in the new scan (if there is no
        # gap between scans, happens for diaPASEF):
        if next_scan_switch == num_scans:
            next_scan_switch = scan_start
            in_scan = True

        mslevel = 2
    elif msms == 8:
        q = conn.execute("SELECT IsolationMz, IsolationWidth, ScanNumBegin, ScanNumEnd, CollisionEnergy FROM PasefFrameMsMsInfo WHERE Frame={0} ORDER BY ScanNumBegin DESC".format(frame_id))
        scandata = q.fetchall()
        tmp = scandata[scan_data_it]
        center = float(tmp[0])
        width = float(tmp[1])
        scan_start = int(tmp[2])
        scan_end = int(tmp[3])
        next_scan_switch = scan_end
        scanBoundariesk0 = td.scanNumToOneOverK0(frame_id, np.array([scan_end, scan_start]))
        # Check if we already are in the new scan (if there is no
        # gap between scans, happens for diaPASEF):
        if next_scan_switch == num_scans:
            next_scan_switch = scan_start
            in_scan = True

        mslevel = 2
    else:
        # MS1 
        q = conn.execute("select NumScans from Frames where Id={0}".format(frame_id))
        tmp = q.fetchone()
        scan_start = 0
        scan_end = int(tmp[0])
        scanBoundariesk0 = td.scanNumToOneOverK0(frame_id, np.array([scan_end, scan_start])) #1/k0 boundaries for ms1 spectrum 
        # print(scanBoundariesk0[1])

    if verbose:
        print("Frame", frame_id, "mslevel", mslevel, msms, "contains nr scans:", num_scans, "and nr pasef scans", len(scandata) if scandata else -1)
        print("Scandata for PASEF:", scandata)
    if keep_frames:
        next_scan_switch = -1

    # Get the mapping of the ion mobility axis
    scan_number_axis = np.arange(num_scans, dtype=np.float64)
    ook0_axis = td.scanNumToOneOverK0(frame_id, scan_number_axis)

    nr_scans_created = 0
    allmz = []
    allint = []
    allim = []

    # Traverse in reversed order to get low ion mobilities first (and high scan times first)
    for k, scan in reversed(list(enumerate(td.readScans(frame_id, 0, num_scans)))):
        index = np.array(scan[0], dtype=np.float64)
        mz = td.indexToMz(frame_id, index)
        intens = scan[1]
        drift_time = ook0_axis [ k ] 
        if compressFrame:
            allmz.append(mz)
            allint.append(intens)
            allim.append( [drift_time for dr_time in mz] )

            # We have multiple MS2 spectra in each frame, we need to separate
            # them based on the information from PasefFrameMsMsInfo which
            # indicates the switch scan and the isolation parameter for each
            # quadrupole isolation.
            if next_scan_switch >= 0 and next_scan_switch >= k:

                if verbose:
                    print("Switch to new scan at", k, "/", next_scan_switch, "store scan of size", len(allmz))

                if in_scan:
                    # Only store spectrum when actually inside a scan, skip the "between scan" pushes
                    sframe = handle_compressed_frame(allmz, allint, allim, mslevel, time, center, width, scanBoundariesk0)
                    sframe.setNativeID("frame=%s_scan=%s" % (frame_id, next_scan_switch) )
                    exp.consumeSpectrum(sframe)
                    nr_scans_created += 1

                allmz = []
                allint = []
                allim = []
                if k == 0: continue

                if in_scan:
                    scan_data_it += 1

                    if scan_data_it >= len(scandata):
                        if verbose: print("LEFT the last scan, nothing else to do here")
                        next_scan_switch = -2
                        continue

                    # Already prepare for next scan
                    tmp = scandata[scan_data_it]
                    center = float(tmp[0])
                    width = float(tmp[1])
                    scan_start = int(tmp[2])
                    scan_end = int(tmp[3])
                    scanBoundariesk0 = td.scanNumToOneOverK0(frame_id, np.array([scan_end, scan_start]))

                    in_scan = False
                    next_scan_switch = scan_end

                    if verbose: print("LEAVING scan now, next scan starts at:", next_scan_switch)

                    # Check if we already are in the new scan (if there is no
                    # gap between scans, happens for diaPASEF):
                    if k == next_scan_switch:
                        if verbose: print("STARTING new scan immediately at", k, ":",  center - width/2.0, center + width/2.0, "scan will end at:", next_scan_switch)
                        next_scan_switch = scan_start
                        in_scan = True
                else:
                    in_scan = True
                    next_scan_switch = scan_start
                    if verbose: print("STARTING new scan at", k, ":",  center - width/2.0, center + width/2.0, "scan will end at:", next_scan_switch)

            continue

        # Store data in OpenMS Spectrum file -> each TOF push is an individual
        # spectrum and we store the ion mobility in the precursor. The frame
        # can be reconstructed by grouping all spectra with the same RT.
        s = pyopenms.MSSpectrum()
        s.setMSLevel(mslevel)
        s.set_peaks( (mz, intens) ) 
        s.setRT(time)
        s.setNativeID("frame=%s spec %s" % (frame_id, k) )
        p = pyopenms.Precursor()
        p.setDriftTime(drift_time)
        if mslevel == 2:
            p.setMZ(center)
            p.setIsolationWindowUpperOffset(width / 2.0)
            p.setIsolationWindowLowerOffset(width / 2.0)
        s.setPrecursors([p])
        exp.consumeSpectrum(s)

    # Store data compressed for cases where the whole frame represents a single spectrum (e.g. MS1)
    if compressFrame and next_scan_switch == -1:
        sframe = handle_compressed_frame(allmz, allint, allim, mslevel, time, center, width, scanBoundariesk0)
        sframe.setNativeID("frame=%s" % frame_id)
        exp.consumeSpectrum(sframe)
        nr_scans_created += 1

    if scandata is not None and (nr_scans_created != len(scandata)):
        raise Exception("Something went quite wrong here, we expected", len(scandata), "scans, but only created", nr_scans_created)

def handle_compressed_frame(allmz, allint, allim, mslevel, rtime, center, width, scanBoundariesk0):
    mz = np.concatenate(allmz)
    intens = np.concatenate(allint)
    ims = np.concatenate(allim)

    fda = pyopenms.FloatDataArray()
    fda.setName("Ion Mobility")
    fda.resize(len(mz))
    for k, val in enumerate(ims):
        fda[k] = val

    sframe = pyopenms.MSSpectrum()
    sframe.setMSLevel(mslevel)
    sframe.setRT(rtime)

    p = pyopenms.Precursor()

    if mslevel == 2:
        p.setMZ(center)
        p.setIsolationWindowUpperOffset(width / 2.0)
        p.setIsolationWindowLowerOffset(width / 2.0)

    # set IM boundaries of spectrum, set here to be consistsent with proteowizard
    sframe.setMetaValue('ion mobility lower limit', scanBoundariesk0[0])
    sframe.setMetaValue('ion mobility upper limit', scanBoundariesk0[1])

    sframe.setPrecursors([p])
    sframe.set_peaks( (mz, intens) )
    sframe.setFloatDataArrays([fda])
    sframe.sortByPosition()

    return sframe

def get_consumer(output_fname):
    # Store output
    if output_fname.lower().endswith("mzml"):
        consumer = pyopenms.PlainMSDataWritingConsumer(output_fname)

        # Compress output
        try:
            opt = consumer.getOptions()
            setCompressionOptions(opt)
            consumer.setOptions(opt)
        except Exception as e:
            print(e)
            print("Your version of pyOpenMS does not support any compression, your files may get rather large")
            pass

    elif output_fname.lower().endswith("sqmass"):
        consumer = pyopenms.MSDataSqlConsumer(output_fname)

    else:
        raise Exception("Supported filenames: mzML and sqMass.")

    return consumer

def convert_diapasef_tdf_to_mzml(analysis_dir, output_fname, merge_scans, keep_frames, verbosity, overlap_scans, frame_limit):
    """
    Main method to perform conversion form bruker tdf data to mzML format
    """

    if sys.version_info.major == 2:
        analysis_dir = unicode(analysis_dir)

    # merge_scans = -1
    # if len(sys.argv) > 3:
    #     merge_scans = int(sys.argv[3])
    #     print("Will merge", merge_scans, "scans")

    td = TimsData(analysis_dir)
    conn = td.conn

    # Get total frame count:
    q = conn.execute("SELECT COUNT(*) FROM Frames")
    row = q.fetchone()
    N = row[0]
    print("Analysis has {0} frames.".format(N))

    consumer = get_consumer(output_fname)

    if merge_scans != -1:
        consumer = MergeConsumer(consumer, merge_scans)

    if overlap_scans > 1:
        
        # For overlapping scans, we need to create N different consumers (N files on disk) 
        # with different file names which will then contain the overlapped spectra 
        
        consumers = []
        for k in range(overlap_scans):
            outspl = output_fname.rsplit(".", 1)
            outname = outspl[0] + "_" + str(k) + "." + outspl[1]
            consumer = get_consumer(outname)
            consumers.append(consumer)

        if merge_scans != -1:
            m_consumers = []
            for c in consumers:
                consumer = MergeConsumer(c, merge_scans)
                m_consumers.append(consumer)
            consumers = m_consumers # use the merge consumers in the overlap consumer

        consumer = SplittingConsumer(consumers)

    if frame_limit[0] > -1 and frame_limit[0] <= N:
        lower_frame = frame_limit[0]
    elif frame_limit[0] == -1:
        lower_frame = 0
    else:
        raise ValueError("Lower Frame limit is not in the permitted range of frames")
    if frame_limit[1] > frame_limit[0] and frame_limit[1] <= N:
        upper_frame = frame_limit[1]
    elif frame_limit[1] == -1:
        upper_frame = N
    else:
        raise ValueError("Upper Frame limit is not in the permitted range of frames")

    for frame_id in tqdm(range(lower_frame, upper_frame)):
        store_frame(frame_id+1, td, conn, consumer, compressFrame=True, verbose=verbosity > 1, keep_frames=keep_frames)
    


def main():

    parser = argparse.ArgumentParser(description ="Conversion program to convert a Bruker raw data file from a timsTOF Pro instrument into a single mzML.")
    parser.add_argument("-a", "--analysis_dir",
                        help = "The location of the directory containing raw data (usually .d)",
                        dest = 'analysis_dir',
                        required = True)
    parser.add_argument("-o", "--output_name",
                        help = "The name of the output file (mzML)",
                        dest = "output_fname",
                        required = True)
    parser.add_argument("-m", "--merge",
                        help = "Number of consecutive frames to sum up (squash). This is useful to boost S/N if exactly repeated frames are measured.",
                        type = int,
                        default = -1,
                        dest = "merge_scans")
    parser.add_argument("--keep_frames",
                        help = "Whether to store frames exactly as measured or split them into individual spectra by precursor isolation window (default is to split them - this is almost always what you want).",
                        type = bool,
                        default = False,
                        dest = "keep_frames")
    parser.add_argument("--verbose",
                        help = "Verbosity",
                        type = int,
                        default = -1,
                        dest = "verbosity")
    parser.add_argument("--overlap",
                        help = "How many overlapping windows were recorded for the same m/z window - will split the output into N output files.",
                        type = int,
                        default = -1,
                        dest = "overlap_scans")
    parser.add_argument("-r", "--framerange",
                        help = "The minimum and maximum Frames to convert. Useful to only convert a part of a file.",
                        type = int,
                        nargs = 2,
                        default = [-1, -1],
                        dest = "frame_limit")
    args = parser.parse_args()
    print("Running conversion with these parameters:\n ")
    analysis_dir = args.analysis_dir
    print("Raw file directory: ", analysis_dir)
    output_fname = args.output_fname
    print("Output name: ", output_fname)
    merge_scans = args.merge_scans
    print("Scans to merge: ", merge_scans)
    overlap_scans = args.overlap_scans
    print("Overlapping scans: ", overlap_scans)
    frame_limit = args.frame_limit
    print("Frame limits: ", frame_limit)

    # if len(sys.argv) < 3:
    #     raise RuntimeError("need arguments: tdf_directory output.mzML")

    # analysis_dir = sys.argv[1]
    # output_fname = sys.argv[2]
    
    convert_diapasef_tdf_to_mzml(analysis_dir, output_fname, merge_scans, args.keep_frames, args.verbosity, overlap_scans, frame_limit)

    print("Conversion completed, press Enter to continue.")

if __name__ == "__main__":
    main()




#!/usr/bin/env python
from __future__ import print_function
from diapysef.timsdata import DiaPasefData
import sys
import pandas as pd
"""
Extracts the SWATH windows from a Bruker raw file.
The resulting windows contain the m/z center and widths and the Ion mobility
start and end coordinates (in scan number and 1/K0 units)
"""
if len(sys.argv) < 3:
    print("Usage: get_dia_windows.py tims_analysis_directory output_file")
    sys.exit()

experiment = sys.argv[1]
outfile = sys.argv[2]

dia = DiaPasefData(experiment)
windows = dia.get_windows()
windows = windows.loc[:, ~windows.columns.duplicated()]

windows.to_csv(outfile)



#!/usr/bin/env python
from __future__ import print_function
import pyopenms
import sys
import os
import platform
import click
import ast

# Logging and performance modules
from functools import wraps
import contextlib
import traceback
from time import time
import logging
from datetime import datetime

# Downloading data
import urllib.request


def setCompressionOptions(opt):
    """
    Adds suitable compression options for an object of type
    pyopenms.PeakFileOptions
        - compresses mass / time arrays with numpress linear
        - compresses intensity with slof (log integer)
        - compresses ion mobility with slof (log integer)
    """
    cfg = pyopenms.NumpressConfig()
    cfg.estimate_fixed_point = True
    cfg.numpressErrorTolerance = -1.0 # skip check, faster
    cfg.setCompression(b"linear");
    cfg.linear_fp_mass_acc = -1; # set the desired RT accuracy in seconds
    opt.setNumpressConfigurationMassTime(cfg)
    cfg = pyopenms.NumpressConfig()
    cfg.estimate_fixed_point = True
    cfg.numpressErrorTolerance = -1.0 # skip check, faster
    cfg.setCompression(b"slof");
    opt.setNumpressConfigurationIntensity(cfg)
    opt.setCompression(True) # zlib compression

    # Now also try to compress float data arrays (this is not enabled in all
    # versions of pyOpenMS).
    try:
        cfg = pyopenms.NumpressConfig()
        cfg.estimate_fixed_point = True
        cfg.numpressErrorTolerance = -1.0 # skip check, faster
        cfg.setCompression(b"slof");
        opt.setNumpressConfigurationFloatDataArray(cfg)
    except Exception:
        pass

def check_bruker_sdk():
    '''
    Check if Bruker SDK exists, otherwise download Bruker SDK from opentims bruker bridge repository
    '''

    if platform.system() == 'Linux':
        if not os.path.exists("libtimsdata.so"):
            url = 'https://raw.githubusercontent.com/MatteoLacki/opentims_bruker_bridge/main/opentims_bruker_bridge/libtimsdata.so'
            click.echo(f"INFO: Could not find Bruker SDK! Attempting to download one from: {url}")
            filename, headers = urllib.request.urlretrieve(url, filename="libtimsdata.so")
    elif platform.system() == 'Windows':
        if not os.path.exists("timsdata.dll"):
            if '64bit' in platform.architecture():
                url = 'https://raw.githubusercontent.com/MatteoLacki/opentims_bruker_bridge/main/opentims_bruker_bridge/win64/timsdata.dll'
                click.echo(f"INFO: Could not find Bruker SDK! Attempting to download one from: {url}")
                filename, headers = urllib.request.urlretrieve(url, filename="timsdata.dll")
            elif '32bit' in platform.architecture():
                url = 'https://raw.githubusercontent.com/MatteoLacki/opentims_bruker_bridge/main/opentims_bruker_bridge/win32/timsdata.dll'
                click.echo(f"INFO: Could not find Bruker SDK! Attempting to download one from: {url}")
                filename, headers = urllib.request.urlretrieve(url, filename="timsdata.dll")

def method_timer(f):
    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        logging.debug('method:%r args:[%r, %r] took: %2.4f sec' % (
            f.__name__, args, kw, te-ts))
        return result
    return wrap


@contextlib.contextmanager
def code_block_timer(ident, log_type=logging.info):
    tstart = time()
    yield
    elapsed = time() - tstart
    log_type("{0}: Elapsed {1} ms".format(ident, elapsed))


def setup_logger(log_file, verbose):
    '''
    Setup logger
    '''
    if verbose == 0:
        use_verbosity = logging.INFO
    else:
        use_verbosity = logging.DEBUG

    root = logging.getLogger(__name__)
    root.setLevel(use_verbosity)
    logging.basicConfig(level=use_verbosity, filename=log_file, filemode='w',
                        format='[%(asctime)s] %(name)s - %(levelname)s - %(message)s')
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(use_verbosity)
    root.addHandler(handler)


def check_sqlite_table(con, table):
    '''
    Check if a table exists in an SQL database
    '''
    table_present = False
    c = con.cursor()
    c.execute(
        'SELECT count(name) FROM sqlite_master WHERE type="table" AND name="%s"' % table)
    if c.fetchone()[0] == 1:
        table_present = True
    else:
        table_present = False
    c.fetchall()

    return (table_present)

def check_im_array(im_array):
    '''
    Check Ion Mobility Array to make sure values are valid
    '''
    
    if any(im_array<0):
      raise click.ClickException(f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] ERROR: There are values below 0 in the input ion mobility array! {im_array[im_array<0]}. Most likely a pyopenms memory view issue.")
    elif 'e' in str(im_array):
      raise click.ClickException(f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] ERROR: The input ion mobility array seems to be in scientific notation! {im_array}. Most likely a pyopenms memory view issue.")

def type_cast_value(value):
    '''
    Convert a string value to python literal
    '''
    if not isinstance(value, str):  # required for Click>=8.0.0
        return value
    try:
        return ast.literal_eval(value)
    except Exception:
        raise click.BadParameter(value)

def argument_value_log(args_dict):
    '''
    Print argument and value
    '''
    logging.debug("---------------- Input Parameters --------------------------")
    for key, item in args_dict.items():
        logging.debug(f"Parameter: {key} = {item}")
    logging.debug("------------------------------------------------------------\n")



#!/usr/bin/env python
from __future__ import print_function
import os, sys
import pandas as pd
import numpy as np
import sqlite3
from diapysef.timsdata import TimsData

class MQData:
    """Read outputs of maxquant for library generation functionalities."""

    def __init__ (self, maxquant_directory):
        if(not os.path.exists(maxquant_directory)):
            raise ValueError("Directory: %s not found. Please make sure that you specified the right directory." % maxquant_directory)
        self.evidence_data = os.path.join(maxquant_directory, "evidence.txt")
        self.msms_data = os.path.join(maxquant_directory, "msms.txt")
        self.all_peptides_data = os.path.join(maxquant_directory, "allPeptides.txt")

    def get_evidence (self):
        """Reads the evidence output of maxquant as pandas dataframe."""
        try:
            self.evidence = pd.read_csv(self.evidence_data, sep = '\t')
        except:
            print("Data File %s not found. Make sure you specified the right directory." % self.evidence_data)

    def get_msms (self):
        """Reads the msms output of maxquant as pandas dataframe."""
        try:
            self.msms = pd.read_csv(self.msms_data, sep = '\t')
        except:
            print("Data File %s not found. Make sure you specified the right directory." % self.msms_data)

    def get_all_peptides (self):
        """Reads the allPeptides output of maxquant as pandas dataframe."""
        try:
            self.all_peptides = pd.read_csv(self.all_peptides_data, sep ='\t')
        except:
            print("Data File %s not found. Make sure you specified the right directory." % self.all_peptides_data)

class PasefMQData(MQData):
    """Reads outputs of maxquant with ion mobility functionalities."""

    def get_evidence (self, timsdata = None):
        """Reads the evidence output of maxquant as pandas dataframe."""
        try:
            self.evidence = pd.read_table(self.evidence_data)
        except:
            print("Data File %s not found. Make sure you specified the right directory." % self.evidence_data)
        if timsdata is not None:
            self.annotate_ion_mobility(timsdata)

    def get_all_peptides (self, timsdata = None):
        """Reads the allPeptides output of maxquant as pandas dataframe."""
        try:
            self.all_peptides = pd.read_table(self.all_peptides_data)
        except:
            print("Data File %s not found. Make sure you specified the right directory." % self.all_peptides_data)
        if timsdata is not None:
            self.annotate_ion_mobility(timsdata)

    def annotate_ion_mobility (self, pasef_data):
        """Adds ion mobility colums to the maxquant results that are currently loaded."""
        if hasattr(self, 'all_peptides'):
            self.all_peptides['IonMobilityIndexK0'] = pasef_data.scanNumToOneOverK0(1,self.all_peptides['Ion mobility index'])

        if hasattr(self, 'evidence'):
            self.evidence['IonMobilityIndexK0'] = pasef_data.scanNumToOneOverK0(1,self.evidence['Ion mobility index'])
    def convert_to_lib(self, irt_file):
        if hasattr(self, 'msms') & hasattr(self, 'evidence'):
            pasef_to_lib(self.evidence, self.msms, irt_file)
        else:
            print("Msms and evidence need to be present for library generation")
            sys.exit()

class PasefData(TimsData):

    # def __init__ (self, analysis_directory):
    #     if sys.version_info.major == 2:
    #         if not isinstance(analysis_directory, unicode):
    #             raise ValueError("analysis_directory must be a Unicode string.")
    #     if sys.version_info.major == 3:
    #         if not isinstance(analysis_directory, str):
    #             raise ValueError("analysis_directory must be a string.")

    #     self.conn = sqlite3.connect(os.path.join(analysis_directory, "analysis.tdf"))

    # def get_conversion_func(self):
    #     q = self.conn.execute("SELECT * FROM TimsCalibration")
    #     calib = q.fetchone()
    #     def convert_im(im):
    #         im = np.array(im)
    #         return(1/(calib[8]+calib[9]/(calib[4]+((calib[5]-calib[4])/calib[3])*(im-calib[6]-calib[2]))))
    #     # Mobility[1/k0] = 1/(c6+c7/(c2+((c3-c2)/c1)*(scanno-c4-c0)))
    #     return convert_im

    # def scanNumToOneOverK0 (self, mzs):
    #     convert_scan_num = self.get_conversion_func()
    #     return(convert_scan_num(mzs))

    def getMetadata(self):
        metadata = pd.read_sql(" SELECT * FROM Frames JOIN PasefFrameMsMsInfo ON Frames.Id = PasefFrameMsMsInfo.Frame JOIN Precursors ON PasefFrameMsMsInfo.Precursor = Precursors.Id" , self.conn)
        self.metadata = metadata

    def filterMetadata(self,
                       imstart=None,
                       imend=None,
                       mzstart=None,
                       mzend=None,
                       rtstart=None,
                       rtend=None,
                       charge=None,
                       metadata=None):
        if metadata is None:
            if self.metadata is None:
                self.getMetadata()
            metadata = self.metadata
            # metadata = pd.read_sql(" SELECT * FROM Frames JOIN PasefFrameMsMsInfo ON Frames.Id = PasefFrameMsMsInfo.Frame JOIN Precursors ON PasefFrameMsMsInfo.Precursor = Precursors.Id" , conn)

        if mzstart is not None:
            scans_filt = metadata[metadata['MonoisotopicMz'] >= mzstart]
        else:
            scans_filt = metadata
        if mzend is not None:
            scans_filt = scans_filt[scans_filt['MonoisotopicMz'] <= mzend]
        if imstart is not None:
            scans_filt = scans_filt[scans_filt['ScanNumBegin'] <= imstart]
        if imend is not None:
            scans_filt = scans_filt[scans_filt['ScanNumEnd'] >= imend]
        if rtstart is not None:
            scans_filt = scans_filt[scans_filt['Time'] >= rtstart]
        if rtend is not None:
            scans_filt = scans_filt[scans_filt['Time'] <= rtend]
        if charge is not None:
            scans_filt = scans_filt[scans_filt['Charge'] == charge]
        return(scans_filt)

        




import sys
import matplotlib
matplotlib.use('Agg')
import pandas as pd
import statsmodels.formula.api as sm
import numpy as np
from statsmodels.graphics.regressionplots import abline_plot
from matplotlib.backends.backend_pdf import PdfPages
from scipy.interpolate import interp1d
import statsmodels.api as smnonlinear

def calibrate(data, plot = True, pdf = "rtcalibration.pdf"):
    """Fits a linear model to the irts"""
    filename = data.iloc[0,0]
    mod = sm.ols(formula = 'irt ~ rt', data = data)
    res = mod.fit()
    # scatter-plot data
    ax = data.plot(x='rt', y='irt', kind='scatter')
    a = abline_plot(model_results=res, ax=ax)
    a = abline_plot(intercept=0, slope=1, ax=ax)
    # print(res.rsquared)
    a.suptitle("%s \n R2: %s \n R2adj: %s" % (filename, res.rsquared, res.rsquared_adj), fontsize = 10)
    # text(0.9, 0.1,("R2:"), ha='center', va='center', transform=ax.transAxes)
    a.savefig(pdf, format = 'pdf')
    matplotlib.pyplot.close()
    return [filename, res.params.Intercept, res.params.rt]

def outliers(data, tolerance = 7):
    '''
    (dataframe, int) -> list of dataframe & list of outliers indices

    Identify outlier indices for lowess alignment.
    Returns a pandas dataframe wwith outliers removed
    '''
    filename = data.iloc[0,0]
    mod = sm.ols(formula = 'irt ~rt', data = data)
    res = mod.fit()
    absd = abs(res.resid) # get the absolute s.d. for each point from linear regression
    outliers = data[absd > 7].index #get the index of the outliers
    data1 = data.drop(outliers)
    return [data1,outliers]

def get_product_charge(msl, msms):
    return(msl)

def align_rt(msms_irt, ms, runs, rt_alignment, pdfout, remove_outliers = True):
    """Retention time alignment"""
    if rt_alignment == 'linear':
        print('Aligning retention time linearly ...')
        # Generate the iRT calibrators
        calibrators = []
        pp = PdfPages(pdfout)
        for file in runs:
            cal = calibrate(msms_irt[msms_irt.raw == file], pdf = pp)
            calibrators.append(cal)
        pp.close()
        # Apply rt calibrators
        calibrators = pd.DataFrame(columns = ['Raw file', 'intercept', 'slope'], data = calibrators)
        ms = pd.merge(ms, calibrators, on = 'Raw file')
        ms['irt'] = ms.intercept + ms.slope * ms['Retention time']
        ms = ms.drop(columns=['intercept', 'slope'])
        
    elif rt_alignment == 'nonlinear':
        print('Aligning retention time with lowess fitting ...')
        lowess = smnonlinear.nonparametric.lowess
        pp = PdfPages(pdfout)
        for file in runs:
            msms_irt_sub = msms_irt[msms_irt.raw == file]
            msms_irt_sub = msms_irt_sub.loc[:,["rt","irt"]]
            if remove_outliers is True:
                msms_irt_sub = outliers(msms_irt_sub)[0]
            # Calculate span
            delta = (max(msms_irt_sub['rt'] - min(msms_irt_sub['rt'])) * 0.01 )
            if len(msms_irt_sub) < 100:
                frac = 1.0
            else:
                frac = 0.1

            # lowess fits by y/x
            r = lowess(msms_irt_sub['irt'], msms_irt_sub['rt'], delta=delta, frac=frac)
            lowess_x = list(zip(*r))[0]
            lowess_y = list(zip(*r))[1]
            # create an interpolation function
            f = interp1d(lowess_x, lowess_y, bounds_error=False)
            nRT = np.asarray(f(ms.loc[ms['Raw file'] == file ,'Retention time'].values))
            
            # fit linear extrapolation for values outside of approximation
            # print(sum(np.isnan(nRT)))
            min_bound = min(lowess_x)
            max_bound = max(lowess_x)

            idx = np.asarray(np.where((ms.loc[ms['Raw file'] == file, 'Retention time'].values < min_bound) | (ms.loc[ms['Raw file'] == file, 'Retention time'].values > max_bound)))   

            lnmod = calibrate(msms_irt_sub, pdf=pp)
            intercept = lnmod[1] # intercept
            slope = lnmod[2] # slope
            nRT[idx] = slope * (ms.loc[ms['Raw file'] == file, 'Retention time'].values[idx]) + intercept
            nrt = []
            for t in nRT: nrt.append(t)
            ms.loc[ms['Raw file'] == file, 'irt'] = list(map(str, nrt))
        pp.close()
    else:
        print('Only rt_alignment: linear and lowess calibrations are currently impletmented.')
        ms['irt'] = ms['Retention time']

    return(ms)


def align_im(msms_irt, ms, runs, im_alignment):
    """Ion mobility calibration."""
    if im_alignment is 'linear':
        print('Aligning ion mobility linearly...')
        # Generate the iIM calibrators
        calibrators = []
        pp = PdfPages('imcalibration.pdf')
        for file in runs:
            msms_irt_sub = msms_irt[msms_irt.raw == file]
            msms_irt_sub = msms_irt_sub.loc[:,["raw","sequence","im","iim"]]
            msms_irt_sub.columns = ["raw","sequence","rt","irt"]
            cal = calibrate(msms_irt_sub, pdf = pp)
            calibrators.append(cal)
        pp.close()
        # Apply im calibrators
        calibrators = pd.DataFrame(columns = ['Raw file', 'intercept', 'slope'], data = calibrators)
        ms = pd.merge(ms, calibrators, on = 'Raw file')
        ms['iim'] = ms.intercept + ms.slope * ms['imcol']
        ms = ms.drop(columns=['intercept', 'slope'])
    else:
        print("Only im_alignment='linear' is currently implemented")
        ms['iim'] = ms['imcol']
    return(ms)

def reformat_mods(data, column):
    data[column] = data[column].str.replace("(ox)","Oxidation")
    data[column] = data[column].str.replace("(ph)","Phospho")
    data[column] = data[column].str.replace("C","C(Carbamidomethyl)")
    data[column] = data[column].str.replace("(ac)","Acetylation")
    data[column] = data[column].str.replace("_\\(ly\\)", "(Label:13C(6)15N(2))")
    data[column] = data[column].str.replace("_\\(ar\\)", "(Label:13C(6)15N(4))")
    data[column] = data[column].str.replace("_", "")
    data[column] = data[column].str.replace("Oxidation \\(M\\)", "Oxidation")
    data[column] = data[column].str.replace("Acetyl \\(Protein N-term\\)", "Acetylation")
      
    
    #data[column] = data[column].str.replace("Oxidation \\(M\\)", "UniMod:35")
    #data[column] = data[column].str.replace("Acetyl \\(Protein N-term\\)", "UniMod:1")
    data[column] = data[column].str.replace("\\(UniMod:4\\)","(Carbamidomethyl)")
    data[column] = data[column].str.replace("\\(UniMod:1\\)","Acetylation")
    data[column] = data[column].str.replace("\\(UniMod:35\\)","Oxidation")
    data[column] = data[column].str.replace("\\(UniMod:21\\)","Phospho")
    data[column] = data[column].str.replace("\\(UniMod:259\\)", "(Label:13C(6)15N(2))")
    data[column] = data[column].str.replace("\\(UniMod:267\\)", "(Label:13C(6)15N(4))")
    return(data)

def pasef_to_tsv(evidence, msms,
                 irt_file = None,
                 ion_mobility = False,
                 pdfout = "rtcalibration.pdf",
                 im_column = 'IonMobilityIndexK0',
                 rt_alignment = 'nonlinear',
                 im_alignment = 'linear'):
    """Converts a mq output to a library taking a best replicate approach."""
    if ion_mobility is True:    
        ev = evidence.loc[:, ["id", "Calibrated retention time", "Ion mobility index", im_column]]
        ev = ev.rename(columns = {im_column:'imcol'})
        ev = ev.rename(columns = {'id':'Evidence ID'})
        ms = pd.merge(msms, ev, on = 'Evidence ID')
        ms = ms.dropna(subset=["Retention time"]) # Some precursors in MQ are not annotated with a RT

        # Replace MQ mods to OpenMS mods
        ms = reformat_mods(ms, "Modified sequence")
        ms = ms.rename(columns = {'Modified sequence':'ModifiedPeptideSequence'})
        
        if irt_file is not None:
            if isinstance(irt_file,str):
               irt = pd.read_table(irt_file)
            elif isinstance(irt_file, pd.DataFrame):
                irt = irt_file
            else:
                print("irt_file must be a path to an irt table or a pd.DataFrame object.")
                sys.exit()
           # Prepare iRT table
            if irt.shape[1] > 2:
                irt_colnames = irt.columns.values.tolist()
                # allowing different formats for reading iRTs
                irt_mod = ['ModifiedPeptideSequence','Modified sequence', 'FullUniModPeptideName', 'FullPeptideName']
                irt_mod = [name for name in irt_colnames if name in irt_mod]
                irt_mod = irt_mod[0]
                irt_rt = ['NormalizedRetentionTime', 'iRT', 'RetentionTime', 'Tr_recalibrated']
                irt_rt = [name for name in irt_colnames if name in irt_rt]
                irt_rt = irt_rt[0]
                irt = irt.loc[:, [irt_mod, irt_rt, "PrecursorIonMobility","PrecursorCharge"]]
                irt = irt.drop_duplicates()
            irt.columns = ["sequence","irt", "iim", "charge"]
            irt = reformat_mods(irt, 'sequence')
            
            # Use the irt peptides present in the MQ output to calibrate
            msms_irt = ms[ms["ModifiedPeptideSequence"].isin(irt["sequence"])]
            ## We chose the best id of each peptide (only one charge state for alignment)
            msms_irt = msms_irt.loc[msms_irt.groupby(["Raw file","ModifiedPeptideSequence"])['PEP'].idxmin()]
            msms_irt = msms_irt.loc[:, ["Raw file","ModifiedPeptideSequence","Retention time", "imcol", "Charge"]] # changed to use raw retention time
            msms_irt.columns = ["raw","sequence","rt","im", "charge"]
            msms_irt = pd.merge(msms_irt, irt, on = ['sequence', 'charge'])
            raw_files = msms_irt.raw.unique()
       
            if rt_alignment is not None:
                ms = align_rt(msms_irt, ms, raw_files, rt_alignment, pdfout, remove_outliers = True)
            else:
                ms['irt'] = ms['Retention time']

            if im_alignment is not None:
                ms = align_im(msms_irt, ms, raw_files, im_alignment)
            else:
                ms['iim'] = ms['imcol']
        else:
            ms['irt'] = ms['Retention time']
            ms['iim'] = ms['imcol']

    else:
        ev = evidence.loc[:, ["id", "Calibrated retention time"]]
        ev = ev.rename(columns = {'id':'Evidence ID'})
        ms = pd.merge(msms, ev, on = 'Evidence ID')
        ms = ms.dropna(subset=["Retention time"]) # Some precursors in MQ are not annotated with a RT

        # Replace MQ mods to OpenMS mods
        ms = reformat_mods(ms, "Modified sequence")
        ms = ms.rename(columns = {'Modified sequence':'ModifiedPeptideSequence'})

        if irt_file is not None:
            if isinstance(irt_file,str):
               irt = pd.read_table(irt_file)
            elif isinstance(irt_file, pd.DataFrame):
                irt = irt_file
            else:
                print("irt_file must be a path to an irt table or a pd.DataFrame object.")
                sys.exit()
           # Prepare iRT table
            if irt.shape[1] > 2:
                irt_colnames = irt.columns.values.tolist()
                # allowing different formats for reading iRTs
                irt_mod = ['ModifiedPeptideSequence','Modified sequence', 'FullUniModPeptideName', 'FullPeptideName']
                irt_mod = [name for name in irt_colnames if name in irt_mod]
                irt_mod = irt_mod[0]
                irt_rt = ['NormalizedRetentionTime', 'iRT', 'RetentionTime', 'Tr_recalibrated']
                irt_rt = [name for name in irt_colnames if name in irt_rt]
                irt_rt = irt_rt[0]
                irt = irt.loc[:, [irt_mod, irt_rt, "PrecursorCharge"]]
                irt = irt.drop_duplicates()
            irt.columns = ["sequence","irt", "charge"]
            irt = reformat_mods(irt, 'sequence')

            # Use the irt peptides present in the MQ output to calibrate
            msms_irt = ms[ms["ModifiedPeptideSequence"].isin(irt["sequence"])]
            ## We chose the best id of each peptide (only one charge state for alignment)
            msms_irt = msms_irt.loc[msms_irt.groupby(["Raw file","ModifiedPeptideSequence"])['PEP'].idxmin()]
            msms_irt = msms_irt.loc[:, ["Raw file","ModifiedPeptideSequence","Retention time", "Charge"]] # changed to raw retention time
            msms_irt.columns = ["raw","sequence","rt", "charge"]
            msms_irt = pd.merge(msms_irt, irt, on = ['sequence', 'charge'])
            raw_files = msms_irt.raw.unique()

            if rt_alignment is not None:
                ms = align_rt(msms_irt, ms, raw_files, rt_alignment, pdfout, remove_outliers = True)
            else:
                ms['irt'] = ms['Retention time']
        else:
            ms['irt'] = ms['Retention time']

    # Filter for best identification (lowest PEP)
    ms = ms.loc[ms.groupby(["ModifiedPeptideSequence", "Charge"])['PEP'].idxmin()]
    # Filter decoy peptides
    ms = ms[ms['Reverse'].isna()]
    
    if ion_mobility is True:
    # Shape data for AssayGenerator
        msl = ms.loc[:, ["id","m/z","Masses","Charge","irt", "iim", "imcol", "Intensities","Sequence","ModifiedPeptideSequence","Proteins"]]
    else:
        msl = ms.loc[:, ["id","m/z","Masses","Charge","irt", "Intensities","Sequence","ModifiedPeptideSequence","Proteins"]]

    masses = msl['Masses'].str.split(';', expand = True).stack().str.strip().reset_index(level = 1, drop=True)
    intensities = msl['Intensities'].str.split(';', expand = True).stack().str.strip().reset_index(level = 1, drop=True)
    df1 = pd.concat([masses, intensities], axis = 1, keys = ['Masses', 'Intensities'])
    msl2 = msl.drop(['Masses', 'Intensities'], axis = 1)
    msl2 = msl2.join(df1).reset_index(drop = True)

    if ion_mobility is True:
        msl2.columns = ["transition_group_id","PrecursorMz","PrecursorCharge","iRT", "Im_recalibrated", "PrecursorIonMobility", "PeptideSequence","FullUniModPeptideName","ProteinName", "ProductMz", "LibraryIntensity"]
    else:
        msl2.columns = ["transition_group_id","PrecursorMz","PrecursorCharge","iRT", "PeptideSequence","FullUniModPeptideName","ProteinName", "ProductMz", "LibraryIntensity"]

    msl2 = get_product_charge(msl2, msms)

    # Reorder the columns as they are in libraries generated with the OSW assay generator
    # msl2 = msl2[["transition_group_id","PrecursorMz","ProductMz","PrecursorCharge","Tr_recalibrated","LibraryIntensity","PeptideSequence","FullUniModPeptideName","ProteinName", "PrecursorIonMobility"]]
    msl2['decoy'] = 0
    msl2['transition_name'] = ['_'.join(str(i) for i in z) for z in zip(msl2.transition_group_id,msl2.PrecursorMz,msl2.ProductMz)]
    msl2=msl2.dropna(subset=['LibraryIntensity'])
    msl2=msl2.drop_duplicates()
    # Write output
    # msl2.to_csv("pasefLib.tsv", sep = "\t", index=False)

    return(msl2)

# @TODO call the OpenSwathAssayGenerator with pyopenms
# def run_assay_generator(paseflib, window_scheme):
#     import pyopenms




from importlib.metadata import metadata
import sys
import os

# MS data modules
import pyopenms as po
from .util import setCompressionOptions, method_timer, code_block_timer, setup_logger, check_sqlite_table, check_im_array, type_cast_value


# Logging and performance modules
import traceback
import logging
import click
from tqdm import tqdm

# Modules for data
import pandas as pd
import numpy as np
import itertools
import sqlite3 as sql3
import pickle as pkl
from joblib import Parallel, delayed, wrap_non_picklable_objects


def generate_coordinates(infile, outfile=None, run_id=None, target_peptides=None, m_score=0.05, use_transition_peptide_mapping=False, use_only_detecting_transitions=True, verbose=0):
    """
    Generate a dictionary of target coordinates to extract from Raw diaPASEF mzML data

    Params:
      infile: a file to generate coordinates from.

    Returns:
      a pickle file containing a dictionary of target coordinates
    """

    if infile.lower().endswith("osw"):
        if verbose == 10:
            logging.debug(
                f"Connecting to OSW identifications database: {infile}")
        con = sql3.connect(infile)

        if use_transition_peptide_mapping:
            use_transition_map = "INNER JOIN TRANSITION_PEPTIDE_MAPPING ON TRANSITION_PEPTIDE_MAPPING.PEPTIDE_ID = PEPTIDE.ID"
            join_on_transition_table_id = "TRANSITION_PEPTIDE_MAPPING.TRANSITION_ID"
        else:
            use_transition_map = "INNER JOIN TRANSITION_PRECURSOR_MAPPING ON TRANSITION_PRECURSOR_MAPPING.PRECURSOR_ID = PRECURSOR.ID"
            join_on_transition_table_id = "TRANSITION_PRECURSOR_MAPPING.TRANSITION_ID"

        if use_only_detecting_transitions:
            detecting_col = '1'
        else:
            detecting_col = '1, 0'

        if check_sqlite_table(con, "SCORE_IPF"):
            join_on_score_table = "INNER JOIN (SELECT * FROM SCORE_IPF) AS SCORE_TABLE ON SCORE_TABLE.FEATURE_ID = FEATURE.ID"
        elif check_sqlite_table(con, "SCORE_MS2"):
            # Restrict to top peak group rank to generate coordinates for. We also filter based on MS2 QVALUE to reduce the number of peptides to generate coordinates for
            join_on_score_table = "INNER JOIN (SELECT * FROM SCORE_MS2 WHERE RANK ==1 AND QVALUE < %s) AS SCORE_TABLE ON SCORE_TABLE.FEATURE_ID = FEATURE.ID" % (m_score)

        # Get Run ID information
        run_ids_df = pd.read_sql_query("SELECT * FROM RUN", con)
        if run_id is None:
            if run_ids_df.shape[0] > 1:
                raise click.ClickException(
                    f"ERROR: Your input osw file {infile} contains more than one run id! Most likely a merged osw, you need to explicilty supply --run_id ID in this case!\n{run_ids_df.to_string()}")
            else:
                run_id = run_ids_df.ID[0]
        else:
            # Check to make sure supplied run id is actually a run in the osw file
            if run_id not in run_ids_df['ID'].to_list():
                raise click.ClickException(
                    f"ERROR: The supplied run_id {run_id} is not a valid run_id in the input osw file {infile}!\n{run_ids_df.to_string()}")

        if target_peptides is not None:

            if os.path.isfile(target_peptides):
                logging.info(
                    f"Reading target peptides from file: {target_peptides}")
                with open(target_peptides, 'r') as file_handle:
                    data = file_handle.read().replace('\n', '')
                target_peptides = [peptide.strip()
                                   for peptide in data.split(',')]
            elif type(target_peptides) == str:
                # If a string list is passed to CLI, then convert to literal python list of strings
                target_peptides = type_cast_value(target_peptides)

            logging.info(
                f"Generating coordinates for the following peptides: {target_peptides}")

            sql_query = '''
      SELECT 
      PEPTIDE.MODIFIED_SEQUENCE AS peptide,
      PRECURSOR.PRECURSOR_MZ AS precursor_mz,
      PRECURSOR.CHARGE AS charge,
      TRANSITION.PRODUCT_MZ AS product_mz,
      TRANSITION.CHARGE AS product_charge,
      TRANSITION.ANNOTATION AS product_annotation,
      TRANSITION.DETECTING AS product_detecting,
      FEATURE.EXP_RT AS rt_apex,
      FEATURE.LEFT_WIDTH AS left_width,
      FEATURE.RIGHT_WIDTH AS right_width,
      FEATURE.EXP_IM AS im_apex,
      SCORE_TABLE.QVALUE as qvalue
      FROM PRECURSOR
      INNER JOIN PRECURSOR_PEPTIDE_MAPPING ON PRECURSOR_PEPTIDE_MAPPING.PRECURSOR_ID = PRECURSOR.ID
      INNER JOIN (
        SELECT * 
        FROM PEPTIDE
        WHERE PEPTIDE.MODIFIED_SEQUENCE IN ("%s")
        ) AS PEPTIDE ON PEPTIDE.ID = PRECURSOR_PEPTIDE_MAPPING.PEPTIDE_ID
      %s 
      INNER JOIN (
        SELECT *
        FROM TRANSITION
        WHERE TRANSITION.DETECTING in (%s)
      ) AS TRANSITION ON TRANSITION.ID = %s
      INNER JOIN (
        SELECT *
        FROM FEATURE
        WHERE FEATURE.RUN_ID=%s
      ) AS FEATURE ON FEATURE.PRECURSOR_ID = PRECURSOR.ID
      %s
      ''' % ('","'.join(target_peptides), use_transition_map, detecting_col, join_on_transition_table_id, run_id, join_on_score_table )

            if verbose == 10:
                logging.debug(f"Injecting SQL Query:\n{sql_query}")

            data = pd.read_sql_query(sql_query, con)
            
            if data.shape[0]==0:
                raise click.ClickException(
                    f"ERROR: Seems like there were none of the target peptides {', '.join(target_peptides)} found in the input osw file {infile}!\n{data.to_string()}")


            data['group_id'] = data['peptide'] + \
                '_' + data['charge'].astype(str)

            # Only keep the best scoring feature per group_id
            data = data.loc[data.groupby('group_id')['qvalue'].transform(
                'min').eq(data['qvalue'])].reset_index(drop=True)

            # Group and aggregate product m/z into a list
            agg_product_mz_df = data.groupby(['group_id'])['product_mz'].apply(
                list).to_frame().reset_index()
            data = pd.merge(data.drop('product_mz', axis=1),
                            agg_product_mz_df, on=['group_id'])
            # Group and aggregate product charge into a list
            agg_product_charge_df = data.groupby(['group_id'])['product_charge'].apply(
                list).to_frame().reset_index()
            data = pd.merge(data.drop('product_charge', axis=1),
                            agg_product_charge_df, on=['group_id'])
            # Group and aggregate product annotation into a list
            agg_product_annotation_df = data.groupby(['group_id'])['product_annotation'].apply(
                list).to_frame().reset_index()
            data = pd.merge(data.drop('product_annotation', axis=1),
                            agg_product_annotation_df, on=['group_id'])
            # Group and aggregate product detecting into a list
            agg_product_detecting_df = data.groupby(['group_id'])['product_detecting'].apply(
                list).to_frame().reset_index()
            data = pd.merge(data.drop('product_detecting', axis=1),
                            agg_product_detecting_df, on=['group_id'])

            # Aggegate left and right bounaries into a list
            data['rt_boundaries'] = data[[
                'left_width', 'right_width']].values.tolist()
            data.drop(['left_width', 'right_width'], axis=1, inplace=True)

            # Melt dataframe to long format
            data_long = data.melt(id_vars=['group_id'])

            # Generate nested dictionary of targeted peptide coordinates
            peptide_coordinates_dict = data_long.groupby(
                'group_id')[['variable', 'value']].apply(lambda x: dict(x.to_numpy())).to_dict()

        # Close connection
        con.close()

    if outfile is not None:
        logging.info(
            f"Writing coordinates dictionary to pickle file {outfile}")
        with open(outfile, "wb") as output_file:
            pkl.dump(peptide_coordinates_dict, file=output_file)
    else:
        return peptide_coordinates_dict


class data_io():
    """
    Class for data input and output operations
    """

    def __init__(self, mzml_file, mz_tol=20, rt_window=50, im_window=0.06, mslevel=[1], readOptions="ondisk", verbose=0, log_file=None):

        # Initialise logger
        if log_file is not None:
            setup_logger(log_file, verbose)

        self.mzml_file = mzml_file
        self.mz_tol = mz_tol
        self.rt_window = rt_window
        self.im_window = im_window
        self.mslevel = mslevel
        self.readOptions = readOptions
        self.verbose = verbose
        self.log_file = log_file
        self.consumer = None

    @method_timer
    def load_data(self, is_filtered=False):
        """
        Method to load data from an mzML file as an on disc experiment for memory efficiency and meta data access without loading full data

        Params:
          is_filtered: (boolean) is the input mzML file a filtered mzML file that is already processed for targeted coordinate extraction
        """

        if is_filtered:
            with code_block_timer('Creating MSExperiment...', logging.debug):
                exp = po.MSExperiment()

            with code_block_timer(f'Loading {self.mzml_file} file...', logging.info):
                po.MzMLFile().load(self.mzml_file, exp)

            logging.info(
                f"There are {exp.getNrSpectra()} spectra and {exp.getNrChromatograms()} chromatograms.")
            logging.info(
                f"There are {sum([spec.getMSLevel()==1  for spec in exp.getSpectra()])} MS1 spectra and {sum([spec.getMSLevel()==2  for spec in exp.getSpectra()])} MS2 spectra.")

            self.filtered = exp
        else:
            if self.readOptions=="ondisk":
                # TODO: For some reason the OnDiscExperiment doesn't read StringMetaData, which is currecntly used to store filtered IM data
                with code_block_timer('Creating OnDiscExperiment...', logging.debug):
                    exp = po.OnDiscMSExperiment()

                with code_block_timer(f'Opening {self.mzml_file} file...', logging.info):
                    exp.openFile(self.mzml_file)

                with code_block_timer(f'Extracting meta data...', logging.debug):
                    meta_data = exp.getMetaData()

                logging.info(
                    f"There are {meta_data.getNrSpectra()} spectra and {exp.getNrChromatograms()} chromatograms.")
                logging.info(
                    f"There are {sum([spec.getMSLevel()==1  for spec in meta_data.getSpectra()])} MS1 spectra and {sum([spec.getMSLevel()==2  for spec in meta_data.getSpectra()])} MS2 spectra.")
            elif self.readOptions=="cached":
                # Because data is cached, we need to make an assumption about whether the file is MS1 or MS2
                with code_block_timer(f'Loading Cached Data from {self.mzml_file} file...', logging.info):
                    exp = po.SpectrumAccessOpenMSCached(self.mzml_file)
                ## Still need access to the meta data, so we need to create an on-disk experiment
                with code_block_timer('Creating OnDiscExperiment...', logging.debug):
                    od_exp = po.OnDiscMSExperiment()
                with code_block_timer(f'Opening {self.mzml_file} file to extract meta-data...', logging.info):
                    od_exp.openFile(self.mzml_file)
                with code_block_timer(f'Extracting meta data...', logging.debug):
                    meta_data = od_exp.getMetaData()
                # Delete OnDiskExperiment since we don't need it anymore, we only need the meta data
                del od_exp
     
                logging.info(
                    f"There are {exp.getNrSpectra()} spectra and {exp.getNrChromatograms()} chromatograms.")
                logging.info(
                    f"There are {sum([spec.getMSLevel()==1  for spec in meta_data.getSpectra()])} MS1 spectra and {sum([spec.getMSLevel()==2  for spec in meta_data.getSpectra()])} MS2 spectra.")
            else:
                click.ClickException(f"ERROR: Unknown readOptions ({self.readOptions}) given! Has to be one of 'ondisk', 'cached'")

            self.exp = exp
            self.meta_data = meta_data
            

    def get_consumer(self, output_fname):
        """
        Get a consumer for mzML or sqMass data writing
        Adpated from diapysef
        """
        # Store output
        if output_fname.lower().endswith("mzml"):
            consumer = po.PlainMSDataWritingConsumer(output_fname)

            # TODO: I've commented out compression, because it seems to make the quality of data bas when uncompressing
            # # Compress output
            # try:
            #     opt = consumer.getOptions()
            #     setCompressionOptions(opt)
            #     consumer.setOptions(opt)
            # except Exception as e:
            #     print(e)
            #     print(
            #         "Your version of pyOpenMS does not support any compression, your files may get rather large")
            #     pass

        elif output_fname.lower().endswith("sqmass"):
            # TODO: For some reason the Sql consumer doesn't work if output file is of type sqmass
            consumer = po.MSDataSqlConsumer(output_fname)

        else:
            raise Exception("Supported filenames: mzML and sqMass.")

        self.consumer = consumer

    @method_timer
    def get_filtered_df(self, ms_level=[1]):
        """
        Get filtered spectrum data into a pandas dataframe format

        Params:
          self: (object) self object that contains filtered data
          ms_level: (list) list of ms level data to write out to file
        """
        results_df = pd.DataFrame()
        for k in tqdm(range(self.filtered.getNrSpectra())):
            spec = self.filtered.getSpectrum(k)
            if spec.getMSLevel() in ms_level:
                mz, intensity = spec.get_peaks()
                rt = np.full([mz.shape[0]], spec.getRT(), float)
                im_tmp = spec.getFloatDataArrays()[0]
                im = im_tmp.get_data()
                add_df = pd.DataFrame({'native_id': spec.getNativeID(), 'ms_level': spec.getMSLevel(
                ), 'peptide': spec.getMetaValue('peptide'), 'mz': mz, 'rt': rt, 'im': im, 'int': intensity})
                results_df = pd.concat([results_df, add_df])
        return results_df

    @method_timer
    def save_filtered_tsv(self, ms_level=[1], out_file="diapasef_extracted_data.tsv"):
        """
        Save a tsv file of the targeted filtered spectra in tabular format

        Params:
          self: (object) self object that contains filtered data
          ms_level: (list) list of ms level data to write out to file
          out_file: (str) output file to write data to
        """
        results_df = pd.DataFrame()
        for k in tqdm(range(self.filtered.getNrSpectra())):
            spec = self.filtered.getSpectrum(k)
            if spec.getMSLevel() in ms_level:
                mz, intensity = spec.get_peaks()
                if (len(mz) == 0 and len(intensity) == 0):
                    logging.warn(
                        f"MS{spec.getMSLevel()} spectrum native id {spec.getNativeID()} had no m/z or intensity array, skipping this spectrum")
                    continue
                rt = np.full([mz.shape[0]], spec.getRT(), float)
                im_tmp = spec.getFloatDataArrays()[0]
                im = im_tmp.get_data()
                precursor = spec.getPrecursors()[0]
                if self.verbose == 10:
                    logging.debug(
                        f"Adding MS{spec.getMSLevel()} spectrum for peptide: {spec.getMetaValue('peptide')} with native id: {spec.getNativeID()}")
                add_df = pd.DataFrame({'native_id': spec.getNativeID(), 'ms_level': spec.getMSLevel(), 'peptide': spec.getMetaValue(
                    'peptide'), 'precursor_mz': precursor.getMZ(), 'charge': precursor.getCharge(), 'mz': mz, 'rt': rt, 'im': im, 'int': intensity})
                results_df = pd.concat([results_df, add_df])
        logging.info(
            f"Saving filtered target spectra data to tabular tsv format: {out_file}")
        results_df.to_csv(out_file, sep="\t")

    def get_target_ms_level_indices(self):
        """
        Extract spectrum indices for a specific mslevel(s).

        Params:
          self: (object) self object containing meta_data

        Return:
          Return a list of indices with request mslevel(s) to self
        """
        if self.readOptions=="ondisk":
            spectra = self.meta_data.getSpectra()
            mslevel_indices = np.array([indice for indice, spec in enumerate(
                spectra) if spec.getMSLevel() in self.mslevel])
            self.mslevel_indices = mslevel_indices
        elif self.readOptions=="cached":
            self.mslevel_indices = np.array(range(0,self.exp.getNrSpectra()))
        else:
            click.ClickException(f"ERROR: Unknown readOptions ({self.readOptions}) given! Has to be one of 'ondisk', 'cached'")

    def get_spectra_rt_list(self):
        """
        Get a list of RT for all the spectra using meta_exp

        Params:
          self: (object) self object containing meta_data

        Return:
          Return a list of RT values for spectra
        """
        if self.readOptions=="ondisk":
            meta_rt_list = np.array([meta_spec.getRT()
                                    for meta_spec in self.meta_data.getSpectra()])
        elif self.readOptions=="cached":
            meta_rt_list = np.array([meta_spec.getRT()
                                    for meta_spec in self.meta_data.getSpectra()])
        else:
            click.ClickException(f"ERROR: Unknown readOptions ({self.readOptions}) given! Has to be one of 'ondisk', 'cached'")
        self.meta_rt_list = meta_rt_list


class TargeteddiaPASEFExperiment(data_io):
    """
    Class for a targeted DIA-PASEF data extraction
    """

    def __init__(self, mzml_file, peptides, mz_tol=20, rt_window=50, im_window=0.06, mslevel=[1], readOptions="ondisk", verbose=0, log_file=None, threads=1):
        """
        Initialize data_access
        """
        super().__init__(mzml_file, mz_tol, rt_window,
                         im_window, mslevel, readOptions, verbose, log_file)
        self.peptides = peptides
        self.threads = threads

    def set_product(self, mz):
        '''
        Create a product container and set the mz
        '''
        product = po.Product()
        product.setMZ(mz)
        return product

    def get_upper_lower_tol(self, target_mz):
        """
        Get the upper bound and lower bound mz around a target mz given a mz tolerance in ppm

        params:
          target_mz: (float) The target mz to generate upper and lower bound mz coordinates for
          mz_tol: (int) The m/z tolerance toget get upper and lower bounds arround target mz. Must be in ppm.

        Return: 
          (tuple) a tuple of the lower and upper bound for target mz
        """
        # mz_uncertainty = target_mz*(self.mz_tol/(2*1000000))
        mz_uncertainty = target_mz*(self.mz_tol/(1000000))
        target_precursor_mz_upper = target_mz + mz_uncertainty
        target_precursor_mz_lower = target_mz - mz_uncertainty
        return target_precursor_mz_lower, target_precursor_mz_upper

    def get_rt_upper_lower(self, rt_apex):
        """
        Get the upper bound and lower bound of a target RT point for a given RT window, i.e. a window of 50 would be 25 points to either side of the target RT.

        params:
          rt_apex: (float) The target RT point to generate upper and lower bound RT coordinates for
          rt_window: (int) The total window range of RT.

        Return:
          (tuple) a tuple of the lower and upper bound for target RT
        """
        return rt_apex-(self.rt_window/2), rt_apex+(self.rt_window/2)

    def get_im_upper_lower(self, im_apex):
        """
        Get the upper bound and lower bound of a target IM point for a given IM window, i.e. a window of 0.06 would be 0.03 points to either side of the target IM.

        params:
          im_apex: (float) The target IM point to generate upper and lower bound IM coordinates for
          im_window: (int) The total window range of IM.

        Return:
          (tuple) a tuple of the lower and upper bound for target IM
        """
        return im_apex-(self.im_window/2), im_apex+(self.im_window/2)

    def is_mz_in_product_mz_tol_window(self, check_mz, target_product_upper_lower_list):
        """
        Check to see if a target product ion mz is within upper and lower tolerance window

        Params: 
          check_mz: (float) The target product ion mz to check
          target_product_upper_lower_list: (list) A list of tuples that contain an upper and lower bound of a product mz

        Return: (bool) Return a logical value 
        """
        return any([check_mz >= bounds[0] and check_mz <= bounds[1] for bounds in target_product_upper_lower_list])

    # @wrap_non_picklable_objects # seems to cause an error with spectrum. list not having attribute getRT()
    def filter_single_spectrum(self, spec_indice, mslevel, target_peptide_group, rt_start, rt_end, im_start, im_end, target_precursor_mz_lower, target_precursor_mz_upper, target_product_upper_lower_list, verbose=0):
        """
        Filter a single spectrum for a given spectrum indice.

        Params:
          self: (TargeteddiaPASEFExperiment object) an object of self that contains an experiment of OnDiskMSExperiment
          spec_indice: (int) an interger of the spectrum indice to extra a spectrum for
          mslevel: (list) a list of intergers for ms levels to filter for
          target_peptide: (str) target peptide sequence string to set as meta data
          rt_start: (float) start of rt window to filter for
          rt_end: (float) end of rt window to filter for 
          im_start: (float) start of ion mobility window to filter for 
          im_end: (float) end of ion mobility window to filter for 
          target_precursor_mz_lower: (float) the lower acceptable bound of the target precursor within m/z tolerance 
          target_precursor_mz_upper: (float) the upper acceptable bound of the target precursor within m/z tolerance 
          target_product_upper_lower_list: (list) a list of tuples containing lower and upper bound for target product m/z within m/z tolerance
          verbose: (int) an interger specifying verbosity level

        Return:
          If self contains a consumer, filtered spectrum is written to disk, otherwise the filtered spectrum MSSpectrum object is retuned
        """
        # Get data arrays
        if (self.readOptions=="ondisk"):
            with code_block_timer(f'Extracting mz, int, im data arrays...', logging.debug):
                spec = self.exp.getSpectrum(spec_indice)
                mz_array = spec.get_peaks()[0]
                int_array = spec.get_peaks()[1]
                im_array = spec.getFloatDataArrays()
        elif (self.readOptions=="cached"):
            spec = self.meta_data.getSpectrum(spec_indice)
            with code_block_timer(f'Extracting mz, int, im data arrays...', logging.debug):
                das = self.exp.getSpectrumById(spec_indice).getDataArrays()
                mz_array = das[0].getData()
                int_array = das[1].getData()
                im_array = das[2].getData()
        else:
            click.ClickException(f"ERROR: Unknown readOptions ({self.readOptions}) given! Has to be one of 'ondisk', 'cached'")
        
        if (self.readOptions=="ondisk"): # TODO: Can we get rid of these annoying if statement checks? I guess we do this because in ondisk exp we need the get_data view
            im_match_bool = (im_array[0].get_data() > im_start) & (
                im_array[0].get_data() < im_end)
        elif (self.readOptions=="cached"):
            im_match_bool = (im_array > im_start) & (
                im_array < im_end)
        
        if spec.getMSLevel() == 1 and 1 in mslevel:
            mz_match_bool = (mz_array > target_precursor_mz_lower) & (
                mz_array < target_precursor_mz_upper)
            if verbose == 10 and any(mz_match_bool*im_match_bool):
                logging.debug(
                    f"Adding MS1 spectrum {spec.getNativeID()} with spectrum indice {spec_indice} filtered for {sum(mz_match_bool*im_match_bool)} spectra between {target_precursor_mz_lower} m/z and {target_precursor_mz_upper} m/z and IM between {im_start} and {im_end}")
        elif spec.getMSLevel() == 2 and 2 in mslevel:
            mz_match_bool = np.array(list(map(self.is_mz_in_product_mz_tol_window, mz_array, itertools.repeat(
                target_product_upper_lower_list, len(mz_array)))))
            if verbose == 10 and any(mz_match_bool*im_match_bool):
                logging.debug(
                    f"Adding MS2 spectrum {spec.getNativeID()} with spectrum indice {spec_indice} filtered for {sum(mz_match_bool*im_match_bool)} spectra between {target_product_upper_lower_list} m/z and IM between {im_start} and {im_end}")
        # Only write out filtered spectra if there is any fitlered spectra to write out
        if any(mz_match_bool*im_match_bool):
            with code_block_timer(f'Getting filtered spectrum...', logging.debug):
                extract_target_indices = np.where(
                    mz_match_bool * im_match_bool)
                filtered_mz = mz_array[extract_target_indices]
                filtered_int = int_array[extract_target_indices]
                if (self.readOptions=="ondisk"):
                    filtered_im = im_array[0].get_data()[extract_target_indices]
                elif (self.readOptions=="cached"):
                    filtered_im = im_array[extract_target_indices]
                # replace peak data with filtered peak data
                spec.set_peaks((filtered_mz, filtered_int))
                # repalce float data arrays with filtered ion mobility data
                fda = po.FloatDataArray()
                filtered_im_np = np.array(filtered_im).astype(np.float32)
                fda.set_data(filtered_im_np)
                fda.setName("Ion Mobility")
                spec.setFloatDataArrays([fda])
                # Set peptide meta data
                spec.setMetaValue(
                    'peptide', self.peptides[target_peptide_group]['peptide'])
                precursor = po.Precursor()
                precursor.setCharge(
                    self.peptides[target_peptide_group]['charge'])
                precursor.setMZ(
                    self.peptides[target_peptide_group]['precursor_mz'])
                spec.setPrecursors([precursor])

                # Set Prodcut mz values
                spec.setProducts([self.set_product(
                    mz, ) for mz in self.peptides[target_peptide_group]['product_mz']])

            # Write out filtered spectra to file if consumer present. This is more memory efficient
            # TODO: Once paralization works, will need to think about how writing to consumer will be affected
            if self.consumer is not None:
                self.consumer.consumeSpectrum(spec)
            else:
                # If you have a lot of filtered spectra to return, it becomes memory heavy.
                return spec

    @method_timer
    def reduce_spectra(self, output_spectra_file=None):
        """
        Main method for filtering raw mzML diaPASEF data given specific set of coordinates to filter for.

        Params:
          self: (TargeteddiaPASEFExperiment object) an object of self that contains an experiment of OnDiskMSExperiment
          output_spectra_file: (str) output mzML file to save filtered spectra to. Recommended for memory efficiency

        Return:
          if an output file is given, filtered data is written to disk, otherwise a filtered MSExperiment object is returned
        """

        # Get Consumer to write filtered spectra to. Memory efficient
        if output_spectra_file is not None:
            logging.debug(
                f"Initializing data consumer to write data to disk - {output_spectra_file}")
            self.get_consumer(output_spectra_file)

        # Get indices for requested ms level spectra
        if self.verbose == 10:
            logging.debug(f"Extracting indices for MS Levels: {self.mslevel}")
        self.get_target_ms_level_indices()

        # Get RT of spectra
        if self.verbose == 10:
            logging.debug(f"Extracting RT values across spectra")
        self.get_spectra_rt_list()
        # logger.info(f"RT Range: {}")

        # TODO: Is this efficient? Or should spectrum access only be extracted during actual spectrum filtering?
        # Main idea behing pre-extracting all desired ms level spectra is to allow for easier paralllization
        # Probably not the best idea to read all spectra into memory. defeats the purpose of on-disk experiment. Will need to find a better way for parallization
        # with code_block_timer(f'Extracting {len(self.mslevel_indices)} of {self.exp.getNrSpectra()} for MS Level {self.mslevel} to spread across {self.threads} threads...'):
        #       spectra_dict = { indice:self.exp.getSpectrum(indice) for indice in self.mslevel_indices.tolist()}

        # Initialize empty MSExperiment container to store filtered data
        if self.consumer is None:
            filtered = po.MSExperiment()
        pbar = tqdm(self.peptides.keys())
        pbar_desc = "INFO: Processing"
        for target_peptide_group in pbar:
            # Update progess bar description
            pbar_desc = f"INFO: Processing..{target_peptide_group}"
            pbar.set_description(pbar_desc)

            # Get Coordinates for current peptide
            target_peptide = self.peptides[target_peptide_group]['peptide']
            target_precursor_mz = self.peptides[target_peptide_group]['precursor_mz']
            target_product_mz = self.peptides[target_peptide_group]['product_mz']
            rt_apex = self.peptides[target_peptide_group]['rt_apex']
            im_apex = self.peptides[target_peptide_group]['im_apex']

            # Get tolerance bounds on coordinates
            target_precursor_mz_lower, target_precursor_mz_upper = self.get_upper_lower_tol(
                target_precursor_mz)
            target_product_upper_lower_list = [
                self.get_upper_lower_tol(mz) for mz in target_product_mz]

            rt_start, rt_end = self.get_rt_upper_lower(rt_apex)
            im_start, im_end = self.get_im_upper_lower(im_apex)

            logging.info(
                f"Extracting data for {target_peptide_group} | target precursor: {target_precursor_mz} m/z ({target_precursor_mz_lower} - {target_precursor_mz_upper}) | RT: {rt_apex} sec ({rt_start} - {rt_end}) | IM: {im_apex} 1/Ko ({im_start} - {im_end})")

            # Restrict spectra list further for RT window, to reduce the number of spectra we need to check to perform filtering on
            use_rt_spec_indices = np.where(np.array(
                [spec_rt >= rt_start and spec_rt <= rt_end for spec_rt in self.meta_rt_list]))
            target_spectra_indices = np.intersect1d(
                self.mslevel_indices, use_rt_spec_indices)
            # spectra_list = [spectra_dict[indice] for indice in np.intersect1d(self.mslevel_indices, use_rt_spec_indices)]

            with code_block_timer(f"Filtering {target_spectra_indices.shape[0]} Spectra for {target_peptide_group}..."):
                try:
                    # TODO: Get prallelization to work.
                    # Current Error is with cython __reduce__
                    # File "/home/justincsing/.local/share/r-miniconda/envs/r-reticulate/lib/python3.6/pickle.py", line 496, in save
                    #     rv = reduce(self.proto)
                    #   File "stringsource", line 2, in pyopenms.pyopenms_1.MSSpectrum.__reduce_cython__
                    # TypeError: self.inst cannot be converted to a Python object for pickling
                    # TODO: Currently very memory heavy, maybe save filtered spectrums to disk? Done.
                    filt_spec_list = Parallel(n_jobs=1)(delayed(self.filter_single_spectrum)(spectrum_indice, self.mslevel, target_peptide_group, rt_start, rt_end, im_start, im_end,
                                                                                             target_precursor_mz_lower, target_precursor_mz_upper, target_product_upper_lower_list, self.verbose) for spectrum_indice in target_spectra_indices.tolist())
                except:
                    traceback.print_exc(file=sys.stdout)

            # Add filtered spectra to MSExperiment container
            if self.consumer is None:
                filt_spec_list = list(
                    filter(lambda spectra: spectra is not None, filt_spec_list))
                # Add filtered spectrum to filtered MSExperiment container
                _ = [filtered.addSpectrum(spec) for spec in filt_spec_list]

        if self.consumer is not None:
            # Close consumer to write final part of data out to file
            del self.consumer
        else:
            self.filtered = filtered



#!/usr/env python

# WRITE A COMMANDLINE TOOL FOR HIGH-PRECISION IRT LIBRARY GENERATION
# FOR THE OPENSWATH tr_irt_nonlinear option

import numpy as np
import re
import pandas as pd
import os
import sys

if len(sys.argv) < 4:
    print("Usage: high_precision_irt.py mqout_dir output_file quality_cutoff")
    print('Description: This script will generate a set of "high precision iRT" peptides as described in Bruderer et al (2016) from MaxQuant output to derive a set of conserved, high-intensity peptides.') 
    sys.exit()

mqout_dir = sys.argv[1]
# mqout_dir can be 1 MQ output from multiple runs or a directory that holds
# multiple MQ out files 

outfile = sys.argv[2]
# One single output high_precision_iRT.txt

cutoff = int(sys.argv[3] )
# Peptide number cutoff

filenames = []
for root, dirs, files in os.walk(mqout_dir):
     for name in files:
         #print(os.path.join(root, name))
         filenames.append(os.path.join(root, name))

r = re.compile(".*msms.txt")

msms_files = list(filter(r.findall, filenames))

if len(msms_files) > 1:
    msms = []
    sequence = []
    for i in msms_files:
        stat = pd.read_csv(i, sep = '\t')
        msms.append(pd.read_csv(i, sep = '\t'))
        sequence.append(stat['Modified sequence'].to_list())

elif len(msms_files) == 1: # check for runs in 1 MQ output
    full_msms = pd.read_csv(msms_files[0], sep = "\t")
    raw_files = full_msms['Raw file'].unique()
    if len(raw_files) > 1: 
        msms = []
        sequence = []
        for k, v in full_msms.groupby('Raw file'):
            df = v
            msms.append(v)
            sequence.append(v['Modified sequence'])
            #print(sequence[0:10])

# get the intersection peptides

intersection_peptides = list(set.intersection(*map(set, sequence)))
print(len(intersection_peptides))


pep = []
for i in msms: pep.append(i['PEP'].mean())

msms_irt = msms[np.argmin(pep)]

msms_irt = msms_irt.loc[msms_irt['Modified sequence'].isin(intersection_peptides)]

msms_irt = msms_irt.sort_values("PEP", ascending = True)
msms_irt = msms_irt.drop_duplicates(subset = "Modified sequence")
msms_irt = msms_irt.rename(columns = {'Modified sequence' : "ModifiedPeptideSequence"})

ev_file = os.path.join(os.path.dirname(msms_files[np.argmin(pep)]), 'evidence.txt')
ev = pd.read_csv(ev_file, sep = '\t')

if 'Calibrated 1/K0' in ev.columns:
    ev = ev.loc[:, ["id", "Calibrated retention time", 'Calibrated 1/K0', 'Ion mobility index']]
    ev = ev.rename(columns = {'id':'Evidence ID', 'Calibrated 1/K0':'PrecursorIonMobility'})
elif 'IonMobilityIndexK0' in ev.columns:
    ev = ev.loc[:, ["id", "Calibrated retention time", "IonMobilityIndexK0"]]
    ev = ev.rename(columns = {'id':'Evidence ID', 'IonMobilityK0':'PrecursorIonMobility'})
else:
    ev = ev.loc[:, ["id", "Calibrated retention time"]]
    ev = ev.rename(columns = {'id': 'Evidence ID'})

ms = pd.merge(msms_irt, ev, on = 'Evidence ID')

rt = ms.loc[:, 'Calibrated retention time']
NormalizedRT = (rt - min(rt)) / (max(rt) - min(rt)) * 100

ms['NormalizedRetentionTime'] = NormalizedRT

ms = ms.sort_values("PEP", ascending = True)

# Filter by PEP (1%) and of these only take the top half of these ...
ms = ms[ms['PEP'] < 0.01]
ms = ms[ms['PEP'] < ms['Score'].quantile(0.5)]

# Then take the number of peptide cutoff as specified
if cutoff is not -1:
    print('Number of peptides: %d' %cutoff)
    ms = ms[0: cutoff]
else:
    print('No cutoff specified, number of peptides: %d' % len(intersection_peptides))


ms1 = ms.loc[:, ["id", "m/z", "Masses", "Charge", "NormalizedRetentionTime","PrecursorIonMobility", "Intensities", "Sequence", "ModifiedPeptideSequence","Proteins"]]

masses = ms1['Masses'].str.split(';', expand =True).stack().str.strip().reset_index(level = 1, drop=True)

intensities = ms1['Intensities'].str.split(';', expand =True).stack().str.strip().reset_index(level = 1, drop=True)

df1 = pd.concat([masses, intensities], axis = 1, keys = ['Masses','Intensities'])

ms12 = ms1.drop(['Masses', 'Intensities'], axis = 1)
ms12 = ms12.join(df1).reset_index(drop = True)
ms12.columns = ['transition_group_id', 'PrecursorMz', 'PrecursorCharge','NormalizedRetentionTime', 'PrecursorIonMobility', 'PeptideSequence','ModifiedPeptideSequence', 'ProteinName', 'ProductMz', 'LibraryIntensity']

ms12['decoy'] = 0
ms12['transition_name'] = ['_'.join(str(i) for i in z) for z in zip(ms12.transition_group_id,ms12.PrecursorMz,ms12.ProductMz)]

ms12=ms12.drop_duplicates()

ms12.to_csv(outfile, sep = "\t", index=False)




import pandas as pd
import os.path

def calculate_coverage (windows, precursor_map = None, count_sc = False):
    """Calculates how many precursors are covered by the window setup."""
    if precursor_map is None:
        mq = pd.read_pickle(os.path.join(os.path.dirname(__file__), 'data/all_peptides.pk'))
    else:
        mq = precursor_map
    if not count_sc:
        mq = mq[mq.Charge > 1]
    w = windows
    w['mzstart'] = w['IsolationMz'] - w['IsolationWidth']/2
    w['mzend'] = w['IsolationMz'] + w['IsolationWidth']/2
    mq['is_covered'] = False
    for i in range(len(windows)):
        mq.loc[(mq['m/z'] < w.loc[i,'mzend']) & (mq['m/z'] > w.loc[i, 'mzstart']) & (mq['IonMobilityIndexK0'] > w.loc[i, 'IMend']) & (mq['IonMobilityIndexK0'] < w.loc[i, 'IMstart']),'is_covered'] = True
    # cov = sum(mq.is_covered)/len(mq)
    return mq



#!/usr/bin/env python3
import ast
import click
import sys
import pickle as pkl
from datetime import datetime

from .util import setup_logger, argument_value_log, check_bruker_sdk
from .targeted_data_extraction import TargeteddiaPASEFExperiment, generate_coordinates
from .convert_tdf_to_mzml import convert_diapasef_tdf_to_mzml
from .plotting import save_report_2d_rt_im_heatmap

# Main Command Line Interface


@click.group(chain=True)
@click.version_option()
def cli():
    '''
    Mobi-DIK (Ion Mobility DIA Tool-Kit) is a package for analysis of DIA data coupled to ion mobility. 

    Visit http://openswath.org/en/latest/docs/mobi-dik.html for usage instructions and help
    '''
    check_bruker_sdk()


# https://stackoverflow.com/a/47730333
class PythonLiteralOption(click.Option):
    def type_cast_value(self, ctx, value):
        if not isinstance(value, str):  # required for Click>=8.0.0
            return value
        try:
            return ast.literal_eval(value)
        except Exception:
            raise click.BadParameter(value)

# Main Targeted DIA-PASEF Data Extraction


@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='Raw data, diaPASEF mzML file.')
@click.option('--coords', 'target_coordinates', required=True, type=click.Path(exists=True), help="""File that contains target coordinates to extract data for, or a file that can be used to generate target coordinates from.\b\n\nCan be one of:\n\npickle (.pkl) - a pickle file that contains a python dictionary of coordinates, i.e. { "TELISVSEVHPS(UniMod:21)R" : {'precursor_mz':767.3691,'charge':2,'product_mz': [311.0639, 312.6357, 322.1867, ..., 11432.6832],'rt_apex':1736.98,'rt_boundaries':[1719.823, 1759.129],'im_apex':0.9884788}, "T(UniMod:21)ELISVSEVHPSR" : {'precursor_mz':767.3691, 'charge':2,'product_mz': [311.0639, 312.6357, 322.1867, ..., 11432.6832],'rt_apex':1730.08,'rt_boundaries':[1718.037, 1751.984],'im_apex':1.0261329}}""")
@click.option('--out', 'outfile', default=('targed_data_extraction.mzML'), show_default=True, type=str, help='Filename to save extracted data to. Must be type mzML')
@click.option('--mz_tol', default=20, show_default=True, type=int, help='The m/z tolerance toget get upper and lower bounds arround target mz. Must be in ppm.')
@click.option('--rt_window', default=50, show_default=True, type=int, help='The total window range of RT, i.e. a window of 50 would be 25 points to either side of the target RT.')
@click.option('--im_window', default=0.06, show_default=True, type=float, help='The total window range of IM, i.e. a window of 0.06 would be 0.03 points to either side of the target IM.')
@click.option('--mslevel', default='[1]', show_default=True, cls=PythonLiteralOption, help='list of mslevel(s) to extract data for. i.e. [1,2] would extract data for MS1 and MS2.')
@click.option('--readOptions', 'readoptions', default='ondisk', show_default=True, type=click.Choice(['ondisk', 'cached']), help='Context to estimate gene-level FDR control.')
@click.option('--verbose', default=0, show_default=True, type=int, help='Level of verbosity. 0 - just displays info, 1 - display some debug info, 10 displays a lot of debug info.')
@click.option('--log_file', default='mobidik_data_extraction.log', show_default=True, type=str, help='Log file to save console messages.')
@click.option('--threads', default=1, show_default=True, type=int, help='Number of threads to parallelize filtering of spectrums across threads.')
def targeted_extraction(infile, target_coordinates, outfile, mz_tol, rt_window, im_window, mslevel, readoptions, verbose, log_file, threads):
    '''
    Extract from the raw data given a set of target coordinates to extract for.
    '''

    # Initialise logger
    setup_logger(log_file, verbose)

    if verbose == 10:
        args_dict = locals()
        argument_value_log(args_dict)

    if target_coordinates.endswith('.pkl'):
        pickle_file = open(target_coordinates, 'rb')
        peptides = pkl.load(pickle_file)
    else:
        raise AssertionError(
            f"Wrong input type ({target_coordinates}) for target coordinates (--coords)! Has to be a pickle file, see --help for example.")

    exp = TargeteddiaPASEFExperiment(
        infile, peptides, mz_tol, rt_window, im_window, mslevel, readoptions, verbose, None, threads)
    click.echo(
        f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] INFO: Loading data...")
    exp.load_data()
    click.echo(
        f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] INFO: Reducing spectra using targeted coordinates...")
    exp.reduce_spectra(outfile)
    click.echo(
        f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] INFO: Finished extracting targeted spectra!")


# Export filtered mzML to tabular data for plotting
@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='A filtered targeted diaPASEF mzML file.')
@click.option('--out', 'outfile', default=('diapasef_extracted_data.tsv'), show_default=True, type=str, help='Filename to save extracted data to.')
@click.option('--mslevel', default='[1]', show_default=True, cls=PythonLiteralOption, help='list of mslevel(s) to extract data for. i.e. [1,2] would extract data for MS1 and MS2.')
@click.option('--verbose', default=0, show_default=True, type=int, help='Level of verbosity. 0 - just displays info, 1 - display some debug info, 10 displays a lot of debug info.')
@click.option('--log_file', default='mobidik_export.log', show_default=True, type=str, help='Log file to save console messages.')
def export(infile, outfile, mslevel, verbose, log_file):
    '''
    Export a reduced targeted mzML file to a tsv file
    '''
    # Initialise logger
    setup_logger(log_file, verbose)

    if verbose == 10:
        args_dict = locals()
        argument_value_log(args_dict)

    exp = TargeteddiaPASEFExperiment(
        infile, None, None, None, None, mslevel, verbose, None, None)
    click.echo(
        f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] INFO: Loading data...")
    exp.load_data(is_filtered=True)
    exp.save_filtered_tsv(mslevel, outfile)
    click.echo(
        f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] INFO: Finished exporting data!")

# Generate a pickle file containing a dictionary of peptide coordinates for targeted data extraction


@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='An OSW file post Pyprophet statiscal validation.')
@click.option('--out', 'outfile', default=('peptides_coordinates.pkl'), show_default=True, type=str, help='Filename to save pickle of peptide coordinates.')
@click.option('--run_id', default=None, show_default=True, type=int, help='Run id in OSW file corresponding to DIA-PASEF run. Required if your OSW file is a merged OSW file.')
@click.option('--target_peptides', default=None, show_default=True, type=str, help='list of peptides (ModifiedPeptideSequence) to generate coordinates for. i.e. ["T(UniMod:21)ELISVSEVHPSR", "TELIS(UniMod:21)VSEVHPSR"]. You can alternatively pass a text file containing comma separate peptide sequences.')
@click.option('--m_score', default=0.05, show_default=True, type=float, help='QValue to filter for peptides below QValue, generate coordinates for these peptides.')
@click.option('--use_transition_peptide_mapping/--no-use_transition_peptide_mapping', default=False, show_default=True, help='Use the TRANSITION_PEPTIDE_MAPPING when getting PRODUCT MZ, instead of joining on TRANSITION_PRECURSOR_MAPPING.')
@click.option('--use_only_detecting_transitions/--no-use_only_detecting_transitions', default=True, show_default=True, help='Only include product m/z of detecting transitions. i.e do not use identifying transitions.')
@click.option('--verbose', default=0, show_default=True, type=int, help='Level of verbosity. 0 - just displays info, 1 - display some debug info, 10 displays a lot of debug info.')
@click.option('--log_file', default='mobidik_peptide_coordinate_generation.log', show_default=True, type=str, help='Log file to save console messages.')
def prepare_coordinates(infile, outfile, run_id, target_peptides, m_score, use_transition_peptide_mapping, use_only_detecting_transitions, verbose, log_file):
    '''
    Generate peptide coordinates for targeted extraction of DIA-PASEF data
    '''
    # Initialise logger
    setup_logger(log_file, verbose)

    if verbose == 10:
        args_dict = locals()
        argument_value_log(args_dict)

    click.echo(
        f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] INFO: Generating coordinates...")
    generate_coordinates(infile, outfile, run_id, target_peptides, m_score,
                         use_transition_peptide_mapping, use_only_detecting_transitions, verbose)
    click.echo(
        f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] INFO: Finished generating coordinates!")

# Conversion program to convert a Bruker TIMS .d data file to mzML format


@cli.command()
@click.option('--in', 'analysis_dir', required=True, type=click.Path(exists=True), help='The location of the directory containing raw data (usually .d).')
@click.option('--out', 'output_fname', required=True, type=str, help='The name of the output file (mzML).')
@click.option('--merge', 'merge_scans', default=-1, show_default=True, type=int, help='Number of consecutive frames to sum up (squash). This is useful to boost S/N if exactly repeated frames are measured.')
@click.option('--keep_frames/--no-keep_frames', 'keep_frames', default=False, show_default=True, help='Whether to store frames exactly as measured or split them into individual spectra by precursor isolation window (default is to split them - this is almost always what you want).')
@click.option('--verbose', 'verbosity', default=-1, show_default=True, type=int, help='Verbosity.')
@click.option('--overlap', 'overlap_scans', default=-1, show_default=True, type=int, help='How many overlapping windows were recorded for the same m/z window - will split the output into N output files.')
@click.option('--framerange', 'frame_limit', default='[-1, -1]', show_default=True, cls=PythonLiteralOption, help='The minimum and maximum Frames to convert. Useful to only convert a part of a file.')
def convertTDFtoMzML(analysis_dir, output_fname, merge_scans, keep_frames, verbosity, overlap_scans, frame_limit):
    '''
    Conversion program to convert a Bruker TIMS .d data file to mzML format
    '''
    click.echo(
        f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] INFO: Converting {analysis_dir}...")
    convert_diapasef_tdf_to_mzml(
        analysis_dir, output_fname, merge_scans, keep_frames, verbosity, overlap_scans, frame_limit)
    click.echo(
        f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] INFO: Finished converting TDF data to mzML!")

# Generate a report for spectic type of plots


@cli.command()
@click.option('--in', 'infile', required=True, type=click.Path(exists=True), help='Data tsv file that contains data to be plotting. i.e peptide sequence, charge state, m/z, MS level, retention time, ion mobility, and intensity')
@click.option('--out', 'outpdf', required=True, type=str, help='The pdf file name to save the plots to.')
@click.option('--type', 'plot_type', default='rt_im_heatmap', show_default=True, type=click.Choice(['rt_im_heatmap']), help='Type of plot to generate.')
# Plot Type: rt_im_heatmap Parameters
@click.option('--plot_contours/--no-plot_contours', 'plot_contours', default=False, show_default=True, help='Should contour lines be plotted? Arg for type rt_im_heatmap')
@click.option('--verbose', default=0, show_default=True, type=int, help='Level of verbosity. 0 - just displays info, 1 - display some debug info, 10 displays a lot of debug info.')
@click.option('--log_file', default='mobidik_report.log', show_default=True, type=str, help='Log file to save console messages.')
def report(infile, outpdf, plot_type, plot_contours, verbose, log_file):
    '''
    Generate a report for a specfific type of plot
    '''
    # Initialise logger
    setup_logger(log_file, verbose)
    if verbose == 10:
        args_dict = locals()
        argument_value_log(args_dict)
    if plot_type == 'rt_im_heatmap':
        click.echo(
            f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] INFO: Generating a report of plots for a Retention Time and Ion Mobility Heatmaps...")
        save_report_2d_rt_im_heatmap(infile, outpdf, plot_contours)
        click.echo(
            f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] INFO: Finished generating report!")
    else:
        raise click.ClickException(f'plot type {plot_type} is not supported')


if __name__ == '__main__':
    cli(obj={})



from setuptools import setup

def readme():
    with open('PYPI_README.rst') as f:
        return f.read()

desc = """\
Analysis, conversion and visualization of diaPASEF data."""

setup(name='diapysef',
      version='1.0.0',
      description=desc,
      long_description=readme(),
      url="https://github.com/Roestlab/dia-pasef",
      author='Max Frank, Annie Ha, Hannes Roest',
      author_email='hannes.rost@utoronto.ca',
      license='MIT',
      packages=['diapysef'],
      install_requires=[
          'pandas',
          'numpy',
          'matplotlib',
          'statsmodels',
          'pyopenms',
          'patsy',
          'tqdm',
          'joblib',
          'click'],
      package_data={
          'diapysef': ['data/*']
      },
      entry_points={
          'console_scripts': [
              "diapysef=diapysef.main:cli",
              ]
      },
      scripts=['scripts/get_dia_windows.py',
               'scripts/annotate_mq_ionmobility.py',
               'scripts/plot_dia_windows.py',
               'scripts/create_library.py',
               'scripts/high_precision_irt.py'],
      zip_safe=False)



#!/usr/bin/env python
from __future__ import print_function

class SplittingConsumer():
    """
        Splitting consumer that deals with splitting up MS2 spectra from the same m/z window.

        This class will distribute incoming spectra into N different downstream
        consumers based on the incoming order and precursor m/z. This can be
        useful when dealing with overlapping windows where spectra need to be
        distributed into alternating files (e.g. the first spectrum with
        precursor 400-425 will go into the first consumer, the second spectrum
        with precursor 400-425 will go into the second consumer, the next one
        into the first consumer again etc.).

        Does not implement chromatogram consuming.
    """

    def __init__(self, consumers):
        self._internal_consumers = consumers
        self._nr_splits = len(consumers)
        self._curr_consumers = {}
        self._cnt = 0

    def __del__(self):
        """
        Cleanup: write all stored spectra to disk
        """
        pass

    def consumeSpectrum(self, s):
        """
            consume individual spectra:
                - write MS1 spectra to all consumers
                - identify current  MS2 spectra and write afer merging n spectra together
        """
        if s.getMSLevel() == 1:
            for c in self._internal_consumers:
                c.consumeSpectrum(s)
        else:

            mz = int(s.getPrecursors()[0].getMZ()*10)
            tmp = self._curr_consumers.get(mz, 0)

            c = self._internal_consumers[tmp]
            c.consumeSpectrum(s)

            self._cnt = self._cnt + 1

            # identify the next consumer to be used
            next_consumer = (tmp+1) % self._nr_splits
            self._curr_consumers[mz] = next_consumer




#!/usr/bin/env python

# Python script to run retention time and ion mobility alignments

#################################################################
#		Import Modules 
################################################################
import diapysef as dp
import pandas as pd
import sys
import os
import click

from diapysef.pasefdata import PasefData
from diapysef.pasefdata import PasefMQData
from diapysef.pasefdata import MQData

###############################################################
#		Load files
##############################################################

@click.command()
#@click.option('--ion_mobility', envvar='ion_mobility', required=True, default=True, help='MaxQuant output contains ion mobility information (set to false for regular MaxQuant output without ion mobility).')
@click.option('--pasefdata', envvar='pasefdata', type=click.Path(exists=True), help='Pasef data file path.')
@click.option('--mqout' , envvar='mqout', required=True, type=click.Path(exists=True), help='Directory from MaxQuant output.')
@click.option('--irt', envvar='irt', type=click.Path(exists=True), help='Full path file of iRT assay library.')
@click.option('--outfile', envvar='outfile', default='mqout.tsv', show_default=True, type=click.Path(exists=False), help='Output assay library file name.')

# Library values alignment
@click.option('--rt_alignment', envvar='rt_alignment', default='nonlinear', show_default=True, type=click.Choice(['linear', 'nonlinear']) , help='Retention time alignment option.')
@click.option('--im_alignment', envvar='im_alignment',default='linear', show_default=True, type=click.Choice(['linear', 'None']), help='Ion Mobility alignment options.')
@click.option('--all_peptides_out', envvar='all_peptides_out', default='annotate1K0_all_peptides.csv', type=click.Path(exists=False), help='Ion mobility annotated all_peptides.csv output file')
@click.option('--evidence_out', envvar='evidence_out', default='annotated1K0_evidence.csv', type=click.Path(exists=False) , help='Ouput evidence.csv file with annotated ion mobility values.')

def main(pasefdata, mqout, irt, outfile, rt_alignment, im_alignment, all_peptides_out, evidence_out):
    mq = MQData(mqout)
    mq.get_evidence()
    ev = mq.evidence
    mq.get_all_peptides()
    all_pep = mq.all_peptides
    if 'Ion mobility index' in ev.columns:
        ion_mobility = True
        mq = PasefMQData(mqout)
        mq.get_all_peptides()
        all_pep = mq.all_peptides
        if 'Calibrated 1/K0' not in ev.columns:
            if pasefdata is None:
                print('WARNING: Ion mobility conversion required. Please specify --pasefdata for ion mobility information')
                sys.exit()
            elif os.path.exists(pasefdata):
                pas = PasefData(pasefdata)
                print('Annotating ion mobility on MQout all_peptides file ...')
                mq.annotate_ion_mobility(pas)
                all_pep = mq.all_peptides
                print('Writing out annotated all_peptides to %s' % all_peptides_out)
                all_pep.to_csv(all_peptides_out)
                mq.get_evidence()
                mq.annotate_ion_mobility(pas)
                print('Annotating ion mobility values for MQout evidence file ...')
                ev = mq.evidence
                print('Writing out annotated evidence file to %s' % evidence_out)
                ev.to_csv(evidence_out)
            else:
                print('WARNING: Ion mobility conversion required. Please specify --pasefdata')
    else:
        ion_mobility = False

    mq.get_msms()
    msms = mq.msms
    pirt = pd.read_csv(irt, sep="\t")

    #############################################################
    #		Perform alignment
    ############################################################
    if ion_mobility is True:
        im_column ='IonMobilityIndexK0'
    else:
        im_column = None
        im_alignment = None
    ptsv = dp.pasef_to_tsv(ev, msms, irt_file=pirt, ion_mobility=ion_mobility, im_column=im_column, rt_alignment=rt_alignment, im_alignment = im_alignment)

    ptsv.to_csv(outfile,sep="\t", index = False)
    print('mqout library generated, proceed to OpenSwathAssayGenerator.')



if __name__ == "__main__":
    main()




#!/usr/bin/env python
from __future__ import print_function
import pandas as pd
import numpy as np
import sqlite3
import os, sys
from ctypes import *

try:
    if sys.platform[:5] == "win32" or sys.platform[:5] == "win64":
        libname = "timsdata.dll"
    elif sys.platform[:5] == "linux":
        libname = "libtimsdata.so"
    else:
        raise Exception("Unsupported platform.")

    dll = cdll.LoadLibrary(os.path.realpath(libname))
    dll.tims_open.argtypes = [ c_char_p, c_uint32 ]
    dll.tims_open.restype = c_uint64
    dll.tims_close.argtypes = [ c_uint64 ]
    dll.tims_close.restype = None
    dll.tims_get_last_error_string.argtypes = [ c_char_p, c_uint32 ]
    dll.tims_get_last_error_string.restype = c_uint32
    dll.tims_has_recalibrated_state.argtypes = [ c_uint64 ]
    dll.tims_has_recalibrated_state.restype = c_uint32
    dll.tims_read_scans_v2.argtypes = [ c_uint64, c_int64, c_uint32, c_uint32, c_void_p, c_uint32 ]
    dll.tims_read_scans_v2.restype = c_uint32

    convfunc_argtypes = [ c_uint64, c_int64, POINTER(c_double), POINTER(c_double), c_uint32 ]

    dll.tims_index_to_mz.argtypes = convfunc_argtypes
    dll.tims_index_to_mz.restype = c_uint32
    dll.tims_mz_to_index.argtypes = convfunc_argtypes
    dll.tims_mz_to_index.restype = c_uint32

    dll.tims_scannum_to_oneoverk0.argtypes = convfunc_argtypes
    dll.tims_scannum_to_oneoverk0.restype = c_uint32
    dll.tims_oneoverk0_to_scannum.argtypes = convfunc_argtypes
    dll.tims_oneoverk0_to_scannum.restype = c_uint32

    dll.tims_scannum_to_voltage.argtypes = convfunc_argtypes
    dll.tims_scannum_to_voltage.restype = c_uint32
    dll.tims_voltage_to_scannum.argtypes = convfunc_argtypes
    dll.tims_voltage_to_scannum.restype = c_uint32

    def throwLastTimsDataError (dll_handle):
        """Throw last TimsData error string as an exception."""

        len = dll_handle.tims_get_last_error_string(None, 0)
        buf = create_string_buffer(len)
        dll_handle.tims_get_last_error_string(buf, len)
        raise RuntimeError(buf.value)

    print("Found Bruker sdk. Access to the raw data is possible. \n")
    sdk = True
except OSError:
    print("Bruker sdk not found. Some functionalities that need access to raw data will not be available. To activate that functionality place libtimsdata.so (Linux) or timsdata.dll in the src folder. \n")
    sdk = False

class TimsData:


    def get_conversion_func(self):
        q = self.conn.execute("SELECT * FROM TimsCalibration")
        calib = q.fetchone()
        def convert_im(frame_id, im):
            im = np.array(im)
            frame_id = frame_id # Currently the conversion does not depend on the frame_id
            return(1/(calib[8]+calib[9]/(calib[4]+((calib[5]-calib[4])/calib[3])*(im-calib[6]-calib[2]))))
        # Mobility[1/k0] = 1/(c6+c7/(c2+((c3-c2)/c1)*(scanno-c4-c0)))
        return convert_im

    if sdk:

        def __init__ (self, analysis_directory, use_recalibrated_state=False):
            if sys.version_info.major == 2:
                if not isinstance(analysis_directory, unicode):
                    raise ValueError("analysis_directory must be a Unicode string.")
            if sys.version_info.major == 3:
                if not isinstance(analysis_directory, str):
                    raise ValueError("analysis_directory must be a string.")

            self.dll = dll

            self.handle = self.dll.tims_open(
                analysis_directory.encode('utf-8'),
                1 if use_recalibrated_state else 0 )
            if self.handle == 0:
                throwLastTimsDataError(self.dll)

            self.conn = sqlite3.connect(os.path.join(analysis_directory, "analysis.tdf"))

            q = self.conn.execute("SELECT MsMsType FROM Frames")

            COLUMN = 0
            frames = q.fetchall()
            ms2_ids = [row[COLUMN] for row in frames]
            self.ms2_id = max(ms2_ids)

            self.initial_frame_buffer_size = 128 # may grow in readScans()
            self.dll = dll

        def __del__ (self):
            if hasattr(self, 'handle'):
                self.dll.tims_close(self.handle)

        def __callConversionFunc (self, frame_id, input_data, func):
            if type(input_data) is np.ndarray and input_data.dtype == np.float64:
                # already "native" format understood by DLL -> avoid extra copy
                in_array = input_data
            else:
                # convert data to format understood by DLL:
                in_array = np.array(input_data, dtype=np.float64)

            cnt = len(in_array)
            out = np.empty(shape=cnt, dtype=np.float64)
            success = func(self.handle, frame_id,
                           in_array.ctypes.data_as(POINTER(c_double)),
                           out.ctypes.data_as(POINTER(c_double)),
                           cnt)
            if success == 0:
                throwLastTimsDataError(self.dll)
            return out

        def indexToMz (self, frame_id, mzs):
            return self.__callConversionFunc(frame_id, mzs, self.dll.tims_index_to_mz)

        def mzToIndex (self, frame_id, mzs):
            return self.__callConversionFunc(frame_id, mzs, self.dll.tims_mz_to_index)

        def scanNumToOneOverK0 (self, frame_id, mzs):
            return self.__callConversionFunc(frame_id, mzs, self.dll.tims_scannum_to_oneoverk0)

        def oneOverK0ToScanNum (self, frame_id, mzs):
            return self.__callConversionFunc(frame_id, mzs, self.dll.tims_oneoverk0_to_scannum)

        def scanNumToVoltage (self, frame_id, mzs):
            return self.__callConversionFunc(frame_id, mzs, self.dll.tims_scannum_to_voltage)

        def voltageToScanNum (self, frame_id, mzs):
            return self.__callConversionFunc(frame_id, mzs, self.dll.tims_voltage_to_scannum)


        # Output: list of tuples (indices, intensities)
        def readScans (self, frame_id, scan_begin, scan_end):
            # buffer-growing loop
            while True:
                cnt = int(self.initial_frame_buffer_size) # necessary cast to run with python 3.5
                buf = np.empty(shape=cnt, dtype=np.uint32)
                len = 4 * cnt

                required_len = self.dll.tims_read_scans_v2(self.handle, frame_id, scan_begin, scan_end,
                                                            buf.ctypes.data_as(POINTER(c_uint32)),
                                                            len)
                if required_len == 0:
                    throwLastTimsDataError(self.dll)

                if required_len > len:
                    if required_len > 16777216:
                        # arbitrary limit for now...
                        raise RuntimeError("Maximum expected frame size exceeded.")
                    self.initial_frame_buffer_size = required_len / 4 + 1 # grow buffer
                else:
                    break

            result = []
            d = scan_end - scan_begin
            for i in range(scan_begin, scan_end):
                npeaks = buf[i-scan_begin]
                indices     = buf[d : d+npeaks]
                d += npeaks
                intensities = buf[d : d+npeaks]
                d += npeaks
                result.append((indices,intensities))

            return result

        def readSpectrum (self, frame_id, scan_begin, scan_end):

            scans = self.readScans(frame_id, scan_begin, scan_end)
            # Summarize on a grid
            allind = []
            allint = []
            for scan in scans:
                indices = np.array(scan[0])
                if len(indices) > 0:
                    # summed_intensities = np.zeros(indices.max() + 1)
                    intens = scan[1]
                    allind = np.concatenate((allind, indices))
                    allint = np.concatenate((allint, intens))
                    # summed_intensities[indices] += intens
            allmz = self.indexToMz(frame_id, allind)
            return((allmz, allint))

    else:

        def __init__ (self, analysis_directory, use_recalibrated_state=False):
            if sys.version_info.major == 2:
                if not isinstance(analysis_directory, unicode):
                    raise ValueError("analysis_directory must be a Unicode string.")
            if sys.version_info.major == 3:
                if not isinstance(analysis_directory, str):
                    raise ValueError("analysis_directory must be a string.")

            self.conn = sqlite3.connect(os.path.join(analysis_directory, "analysis.tdf"))

        def scanNumToOneOverK0 (self, frame_id, mzs):
            convert_scan_num = self.get_conversion_func()
            return(convert_scan_num(frame_id, mzs))

class DiaPasefData(TimsData):

    def get_cycle_length (self):
        firstcycle = True
        initial_ms1 = True
        cycle_length = 0
        row = 1
        while (firstcycle):
            mstype = self.conn.execute("SELECT MsMsType FROM Frames LIMIT {0},1".format(row)).fetchone()
            row += 1
            if mstype[0] == self.ms2_id:
                cycle_length += 1
                initial_ms1 = False
            elif(initial_ms1 != True):
                firstcycle = False
        return cycle_length

    def get_windows_per_frame(self):
        if self.ms2_id == 8:
            q = self.conn.execute("SELECT * FROM Frames INNER JOIN PasefFrameMsMsInfo ON Frames.Id=PasefFrameMsMsInfo.Frame WHERE MsMsType={0} AND Frame = (SELECT MIN(Frame) FROM PasefFrameMsMsInfo as fr)".format(self.ms2_id))
        elif self.ms2_id == 9:
            q = self.conn.execute("SELECT * FROM Frames INNER JOIN DiaFrameMsMsInfo ON Frames.Id = DiaFrameMsMsInfo.Frame INNER JOIN DiaFrameMsMsWindows ON DiaFrameMsMsInfo.WindowGroup = DiaFrameMsMsWindows.WindowGroup WHERE MsMsType={0} AND Frame = (SELECT MIN(Frame) FROM DiaFrameMsMsInfo as fr)".format(self.ms2_id))
        windows = q.fetchall()
        return len(windows)

    def get_windows(self):
        """Extracts the window scheme from the first cycle of a tims file"""
        #pasef_ms2_id = 8 # diaPASEF ms2 scans are denoted by 8 instead of 2
        if self.ms2_id == 8:
            cycle_length = self.get_cycle_length()
            wpf = self.get_windows_per_frame()
            q = self.conn.execute("SELECT * FROM Frames INNER JOIN PasefFrameMsMsInfo ON Frames.Id=PasefFrameMsMsInfo.Frame WHERE MsMsType={0} LIMIT {1}" .format(self.ms2_id, cycle_length*wpf))
        elif self.ms2_id == 9:
            cycle_length = self.get_cycle_length()
            wpf = self.get_windows_per_frame()
            q = self.conn.execute("SELECT * FROM Frames INNER JOIN DiaFrameMsMsInfo ON Frames.Id = DiaFrameMsMsInfo.Frame INNER JOIN DiaFrameMsMsWindows ON DiaFrameMsMsInfo.WindowGroup = DiaFrameMsMSWindows.WindowGroup WHERE MsMsType={0} LIMIT {1}".format(self.ms2_id, cycle_length*wpf))
        frames = q.fetchall()
        colnames = [description[0] for description in q.description]
        resframe = pd.DataFrame(data = frames, columns = colnames)
        # Currently the frame does not matter for the conversion, thus we can set it to 1
        resframe["IMstart"] = self.scanNumToOneOverK0(1, resframe.ScanNumBegin)
        resframe["IMend"] = self.scanNumToOneOverK0(1, resframe.ScanNumEnd)
        return resframe
